import os

import numpy as np
from numpy.testing import assert_array_equal, assert_equal

import yt.units.dimensions as dimensions
from yt.geometry.oct_container import _ORDER_MAX
from yt.geometry.particle_oct_container import ParticleBitmap, ParticleOctreeContainer
from yt.geometry.selection_routines import RegionSelector
from yt.testing import requires_module
from yt.units.unit_registry import UnitRegistry
from yt.units.yt_array import YTArray
from yt.utilities.lib.geometry_utils import (
    get_hilbert_indices,
    get_hilbert_points,
    get_morton_indices,
    get_morton_points,
)

NPART = 32**3
DLE = np.array([0.0, 0.0, 0.0])
DRE = np.array([10.0, 10.0, 10.0])
DW = DRE - DLE
PER = np.array([0, 0, 0], "bool")
dx = DW / (2**_ORDER_MAX)


def test_add_particles_random():
    np.random.seed(0x4D3D3D3)
    pos = np.random.normal(0.5, scale=0.05, size=(NPART, 3)) * (DRE - DLE) + DLE
    # Now convert to integers
    for i in range(3):
        np.clip(pos[:, i], DLE[i], DRE[i], pos[:, i])
    # Convert to integers
    pos = np.floor((pos - DLE) / dx).astype("uint64")
    morton = get_morton_indices(pos)
    morton.sort()
    for ndom in [1, 2, 4, 8]:
        octree = ParticleOctreeContainer((1, 1, 1), DLE, DRE)
        octree.n_ref = 32
        for split in np.array_split(morton, ndom):
            octree.add(split)
        octree.finalize()
        # This visits every oct.
        tc = octree.recursively_count()
        total_count = np.zeros(len(tc), dtype="int32")
        for i in sorted(tc):
            total_count[i] = tc[i]
        assert_equal(octree.nocts, total_count.sum())
        # This visits every cell -- including those covered by octs.
        # for dom in range(ndom):
        #    level_count += octree.count_levels(total_count.size-1, dom, mask)
        assert_equal(total_count, [1, 8, 64, 64, 256, 536, 1856, 1672])


class FakeDS:
    domain_left_edge = None
    domain_right_edge = None
    domain_width = None
    unit_registry = UnitRegistry()
    unit_registry.add("code_length", 1.0, dimensions.length)
    periodicity = (False, False, False)


class FakeRegion:
    def __init__(self, nfiles, periodic=False):
        self.ds = FakeDS()
        self.ds.domain_left_edge = YTArray(
            [0.0, 0.0, 0.0], "code_length", registry=self.ds.unit_registry
        )
        self.ds.domain_right_edge = YTArray(
            [nfiles, nfiles, nfiles],
            "code_length",
            registry=self.ds.unit_registry,
            dtype="float64",
        )
        self.ds.domain_width = self.ds.domain_right_edge - self.ds.domain_left_edge
        self.ds.periodicity = (periodic, periodic, periodic)
        self.nfiles = nfiles

    def set_edges(self, file_id, dx=0.1):
        self.left_edge = YTArray(
            [file_id + dx, 0.0, 0.0], "code_length", registry=self.ds.unit_registry
        )
        self.right_edge = YTArray(
            [file_id + 1 - dx, self.nfiles, self.nfiles],
            "code_length",
            registry=self.ds.unit_registry,
        )


class FakeBoxRegion:
    def __init__(self, nfiles, left_edge, right_edge):
        self.ds = FakeDS()
        self.ds.domain_left_edge = YTArray(
            left_edge, "code_length", registry=self.ds.unit_registry
        )
        self.ds.domain_right_edge = YTArray(
            right_edge, "code_length", registry=self.ds.unit_registry
        )
        self.ds.domain_width = self.ds.domain_right_edge - self.ds.domain_left_edge
        self.nfiles = nfiles

    def set_edges(self, center, width):
        self.left_edge = self.ds.domain_left_edge + self.ds.domain_width * (
            center - width / 2
        )
        self.right_edge = self.ds.domain_left_edge + self.ds.domain_width * (
            center + width / 2
        )


def FakeBitmap(
    npart,
    nfiles,
    order1,
    order2,
    left_edge=None,
    right_edge=None,
    periodicity=None,
    decomp="sliced",
    buff=0.1,
    distrib="uniform",
    fname=None,
):
    if left_edge is None:
        left_edge = np.array([0.0, 0.0, 0.0])
    if right_edge is None:
        right_edge = np.array([1.0, 1.0, 1.0])
    if periodicity is None:
        periodicity = np.array([0, 0, 0], "bool")
    reg = ParticleBitmap(
        left_edge, right_edge, periodicity, 12345, nfiles, order1, order2
    )
    # Load from file if it exists
    if isinstance(fname, str) and os.path.isfile(fname):
        reg.load_bitmasks(fname)
    else:
        # Create positions for each file
        posgen = yield_fake_decomp(
            decomp, npart, nfiles, left_edge, right_edge, buff=buff, distrib=distrib
        )
        # Coarse index
        max_npart = 0
        for i, (pos, hsml) in enumerate(posgen):
            max_npart = max(max_npart, pos.shape[0])
            reg._coarse_index_data_file(pos, hsml, i)
            reg._set_coarse_index_data_file(i)
        if i != (nfiles - 1):
            raise RuntimeError(
                f"There are positions for {i + 1} files, but there should be {nfiles}."
            )
        # Refined index
        mask = reg.masks.sum(axis=1).astype("uint8")
        sub_mi1 = np.zeros(max_npart, "uint64")
        sub_mi2 = np.zeros(max_npart, "uint64")
        posgen = yield_fake_decomp(
            decomp, npart, nfiles, left_edge, right_edge, buff=buff, distrib=distrib
        )
        coll = None
        for i, (pos, hsml) in enumerate(posgen):
            nsub_mi, coll = reg._refined_index_data_file(
                coll,
                pos,
                hsml,
                mask,
                sub_mi1,
                sub_mi2,
                i,
                0,
                count_threshold=1,
                mask_threshold=2,
            )
            reg.bitmasks.append(i, coll)
        # Save if file name provided
        if isinstance(fname, str):
            reg.save_bitmasks(fname)
    return reg


def test_bitmap_no_collisions():
    # Test init for slabs of points in x
    left_edge = np.array([0.0, 0.0, 0.0])
    right_edge = np.array([1.0, 1.0, 1.0])
    periodicity = np.array([0, 0, 0], "bool")
    npart = 100
    nfiles = 2
    file_hash = 12345
    order1 = 2
    order2 = 2
    reg = ParticleBitmap(
        left_edge, right_edge, periodicity, file_hash, nfiles, order1, order2
    )
    # Coarse index
    posgen = yield_fake_decomp("sliced", npart, nfiles, left_edge, right_edge)
    max_npart = 0
    for i, (pos, hsml) in enumerate(posgen):
        reg._coarse_index_data_file(pos, hsml, i)
        max_npart = max(max_npart, pos.shape[0])
        reg._set_coarse_index_data_file(i)
        assert_equal(reg.count_total(i), np.sum(reg.masks[:, i]))
    mask = reg.masks.sum(axis=1).astype("uint8")
    ncoll = np.sum(mask > 1)
    nc, nm = reg.find_collisions_coarse()
    assert_equal(nc, 0, "%d coarse collisions" % nc)
    assert_equal(ncoll, nc, "%d in mask, %d in bitmap" % (ncoll, nc))
    # Refined index
    sub_mi1 = np.zeros(max_npart, "uint64")
    sub_mi2 = np.zeros(max_npart, "uint64")
    posgen = yield_fake_decomp("sliced", npart, nfiles, left_edge, right_edge)
    coll = None
    for i, (pos, hsml) in enumerate(posgen):
        nsub_mi, coll = reg._refined_index_data_file(
            coll,
            pos,
            hsml,
            mask,
            sub_mi1,
            sub_mi2,
            i,
            0,
            count_threshold=1,
            mask_threshold=2,
        )
        reg.bitmasks.append(i, coll)
        assert_equal(reg.count_refined(i), 0)
    nr, nm = reg.find_collisions_refined()
    assert_equal(nr, 0, "%d collisions" % nr)


def test_bitmap_collisions():
    # Test init for slabs of points in x
    left_edge = np.array([0.0, 0.0, 0.0])
    right_edge = np.array([1.0, 1.0, 1.0])
    periodicity = np.array([0, 0, 0], "bool")
    nfiles = 2
    file_hash = 12345
    order1 = 2
    order2 = 2
    reg = ParticleBitmap(
        left_edge, right_edge, periodicity, file_hash, nfiles, order1, order2
    )
    # Use same points for all files to force collisions
    pos = cell_centers(order1 + order2, left_edge, right_edge)
    hsml = None
    # Coarse index
    max_npart = 0
    for i in range(nfiles):
        reg._coarse_index_data_file(pos, hsml, i)
        max_npart = max(max_npart, pos.shape[0])
        reg._set_coarse_index_data_file(i)
        assert_equal(reg.count_total(i), np.sum(reg.masks[:, i]))
    mask = reg.masks.sum(axis=1).astype("uint8")
    ncoll = np.sum(mask > 1)
    nc, nm = reg.find_collisions_coarse()
    assert_equal(ncoll, nc, "%d in mask, %d in bitmap" % (ncoll, nc))
    assert_equal(nc, 2 ** (3 * order1), "%d coarse collisions" % nc)
    # Refined index
    sub_mi1 = np.zeros(max_npart, "uint64")
    sub_mi2 = np.zeros(max_npart, "uint64")
    for i in range(nfiles):
        nsub_mi, coll = reg._refined_index_data_file(
            None,
            pos,
            hsml,
            mask,
            sub_mi1,
            sub_mi2,
            i,
            0,
            count_threshold=1,
            mask_threshold=2,
        )
        reg.bitmasks.append(i, coll)
        assert_equal(reg.count_refined(i), ncoll)
    nr, nm = reg.find_collisions_refined()
    assert_equal(nr, 2 ** (3 * (order1 + order2)), "%d collisions" % nr)


@requires_module("h5py")
def test_bitmap_save_load():
    # Test init for slabs of points in x
    left_edge = np.array([0.0, 0.0, 0.0])
    right_edge = np.array([1.0, 1.0, 1.0])
    periodicity = np.array([0, 0, 0], "bool")
    npart = NPART
    file_hash = 12345
    nfiles = 32
    order1 = 2
    order2 = 2
    fname_fmt = "temp_bitmasks{}.dat"
    i = 0
    fname = fname_fmt.format(i)
    while os.path.isfile(fname):
        i += 1
        fname = fname_fmt.format(i)
    # Create bitmap and save to file
    reg0 = FakeBitmap(npart, nfiles, order1, order2, left_edge, right_edge, periodicity)
    reg0.save_bitmasks(fname, 0.0)
    # Attempt to load bitmap
    reg1 = ParticleBitmap(
        left_edge, right_edge, periodicity, file_hash, nfiles, order1, order2
    )
    reg1.load_bitmasks(fname)
    assert reg0.iseq_bitmask(reg1)
    # Remove file
    os.remove(fname)


def test_bitmap_select():
    np.random.seed(0x4D3D3D3)
    dx = 0.1
    for periodic in [False, True]:
        for nfiles in [2, 15, 31, 32, 33]:
            # Now we create particles
            # Note: we set order1 to log2(nfiles) here for testing purposes to
            # ensure no collisions
            order1 = int(np.ceil(np.log2(nfiles)))  # Ensures zero collisions
            order2 = 2  # No overlap for N = nfiles
            exact_division = nfiles == (1 << order1)
            div = float(nfiles) / float(1 << order1)
            reg = FakeBitmap(
                nfiles**3,
                nfiles,
                order1,
                order2,
                decomp="grid",
                left_edge=np.array([0.0, 0.0, 0.0]),
                right_edge=np.array([nfiles, nfiles, nfiles]),
                periodicity=np.array([periodic, periodic, periodic]),
            )
            # Loop over regions selecting single files
            fr = FakeRegion(nfiles, periodic=periodic)
            for i in range(nfiles):
                fr.set_edges(i, dx)
                selector = RegionSelector(fr)
                (df, gf), (dmask, gmask) = reg.identify_data_files(selector, ngz=1)
                if exact_division:
                    assert_equal(len(df), 1, f"selector {i}, number of files")
                    assert_equal(df[0], i, f"selector {i}, file selected")
                    if periodic and (nfiles != 2):
                        ans_gf = sorted([(i - 1) % nfiles, (i + 1) % nfiles])
                    elif i == 0:
                        ans_gf = [i + 1]
                    elif i == (nfiles - 1):
                        ans_gf = [i - 1]
                    else:
                        ans_gf = [i - 1, i + 1]
                    assert_equal(
                        len(gf),
                        len(ans_gf),
                        f"selector {i}, number of ghost files",
                    )
                    for i in range(len(gf)):
                        assert_equal(gf[i], ans_gf[i], f"selector {i}, ghost files")

                else:
                    lf_frac = np.floor(float(fr.left_edge[0]) / div) * div
                    rf_frac = np.floor(float(fr.right_edge[0]) / div) * div
                    # Selected files
                    lf = int(
                        np.floor(lf_frac)
                        if ((lf_frac % 0.5) == 0)
                        else np.round(lf_frac)
                    )
                    rf = int(
                        np.floor(rf_frac)
                        if ((rf_frac % 0.5) == 0)
                        else np.round(rf_frac)
                    )
                    if (rf + 0.5) >= (rf_frac + div):
                        rf -= 1
                    if (lf + 0.5) <= (lf_frac - div):
                        lf += 1
                    df_ans = np.arange(max(lf, 0), min(rf + 1, nfiles))
                    assert_array_equal(df, df_ans, f"selector {i}, file array")
                    # Ghost zones selected files
                    lf_ghost = int(
                        np.floor(lf_frac - div)
                        if (((lf_frac - div) % 0.5) == 0)
                        else np.round(lf_frac - div)
                    )
                    rf_ghost = int(
                        np.floor(rf_frac + div)
                        if (((rf_frac + div) % 0.5) == 0)
                        else np.round(rf_frac + div)
                    )
                    if not periodic:
                        lf_ghost = max(lf_ghost, 0)
                        rf_ghost = min(rf_ghost, nfiles - 1)
                    if (rf_ghost + 0.5) >= (rf_frac + 2 * div):
                        rf_ghost -= 1
                    gf_ans = []
                    if lf_ghost < lf:
                        gf_ans.append(lf_ghost % nfiles)
                    if rf_ghost > rf:
                        gf_ans.append(rf_ghost % nfiles)
                    gf_ans = np.array(sorted(gf_ans))
                    assert_array_equal(gf, gf_ans, f"selector {i}, ghost file array")


def cell_centers(order, left_edge, right_edge):
    ndim = left_edge.size
    ncells = 2**order
    dx = (right_edge - left_edge) / (2 * ncells)
    d = [
        np.linspace(left_edge[i] + dx[i], right_edge[i] - dx[i], ncells)
        for i in range(ndim)
    ]
    dd = np.meshgrid(*d)
    return np.vstack([x.flatten() for x in dd]).T


def fake_decomp_random(npart, nfiles, ifile, DLE, DRE, buff=0.0):
    np.random.seed(0x4D3D3D3 + ifile)
    nPF = int(npart / nfiles)
    nR = npart % nfiles
    if ifile == 0:
        nPF += nR
    pos = np.empty((nPF, 3), "float64")
    for i in range(3):
        pos[:, i] = np.random.uniform(DLE[i], DRE[i], nPF)
    return pos


def fake_decomp_sliced(npart, nfiles, ifile, DLE, DRE, buff=0.0):
    np.random.seed(0x4D3D3D3 + ifile)
    DW = DRE - DLE
    div = DW / nfiles
    nPF = int(npart / nfiles)
    nR = npart % nfiles
    inp = nPF
    if ifile == 0:
        inp += nR
    iLE = DLE[0] + ifile * div[0]
    iRE = iLE + div[0]
    if ifile != 0:
        iLE -= buff * div[0]
    if ifile != (nfiles - 1):
        iRE += buff * div[0]
    pos = np.empty((inp, 3), dtype="float")
    pos[:, 0] = np.random.uniform(iLE, iRE, inp)
    for i in range(1, 3):
        pos[:, i] = np.random.uniform(DLE[i], DRE[i], inp)
    return pos


def makeall_decomp_hilbert_gaussian(
    npart,
    nfiles,
    DLE,
    DRE,
    buff=0.0,
    order=6,
    verbose=False,
    fname_base=None,
    nchunk=10,
    width=None,
    center=None,
    frac_random=0.1,
):
    import pickle

    np.random.seed(0x4D3D3D3)
    DW = DRE - DLE
    if fname_base is None:
        fname_base = f"hilbert{order}_gaussian_np{npart}_nf{nfiles}_"
    if width is None:
        width = 0.1 * DW
    if center is None:
        center = DLE + 0.5 * DW

    def load_pos(file_id):
        filename = fname_base + f"file{file_id}"
        if os.path.isfile(filename):
            fd = open(filename, "rb")
            positions = pickle.load(fd)
            fd.close()
        else:
            positions = np.empty((0, 3), dtype="float64")
        return positions

    def save_pos(file_id, positions):
        filename = fname_base + f"file{file_id}"
        fd = open(filename, "wb")
        pickle.dump(positions, fd)
        fd.close()

    npart_rnd = int(frac_random * npart)
    npart_gau = npart - npart_rnd
    dim_hilbert = 1 << order
    nH = dim_hilbert**3
    if nH < nfiles:
        raise ValueError("Fewer hilbert cells than files.")
    nHPF = nH / nfiles
    rHPF = nH % nfiles
    for ichunk in range(nchunk):
        inp = npart_gau / nchunk
        if ichunk == 0:
            inp += npart_gau % nchunk
        pos = np.empty((inp, 3), dtype="float64")
        ind = np.empty((inp, 3), dtype="int64")
        for k in range(3):
            pos[:, k] = np.clip(
                np.random.normal(center[k], width[k], inp),
                DLE[k],
                DRE[k] - (1.0e-9) * DW[k],
            )
            ind[:, k] = (pos[:, k] - DLE[k]) / (DW[k] / dim_hilbert)
        harr = get_hilbert_indices(order, ind)
        farr = (harr - rHPF) / nHPF
        for ifile in range(nfiles):
            ipos = load_pos(ifile)
            if ifile == 0:
                idx = farr <= ifile  # Put remainders in first file
            else:
                idx = farr == ifile
            ipos = np.concatenate((ipos, pos[idx, :]), axis=0)
            save_pos(ifile, ipos)
    # Random
    for ifile in range(nfiles):
        ipos = load_pos(ifile)
        ipos_rnd = fake_decomp_hilbert_uniform(
            npart_rnd, nfiles, ifile, DLE, DRE, buff=buff, order=order, verbose=verbose
        )
        ipos = np.concatenate((ipos, ipos_rnd), axis=0)
        save_pos(ifile, ipos)


def fake_decomp_hilbert_gaussian(
    npart, nfiles, ifile, DLE, DRE, buff=0.0, order=6, verbose=False, fname=None
):
    np.random.seed(0x4D3D3D3)
    DW = DRE - DLE
    dim_hilbert = 1 << order
    nH = dim_hilbert**3
    if nH < nfiles:
        raise Exception("Fewer hilbert cells than files.")
    nHPF = nH / nfiles
    rHPF = nH % nfiles
    hdiv = DW / dim_hilbert
    if ifile == 0:
        hlist = np.arange(0, nHPF + rHPF, dtype="int64")
    else:
        hlist = np.arange(ifile * nHPF + rHPF, (ifile + 1) * nHPF + rHPF, dtype="int64")
    hpos = get_hilbert_points(order, hlist)
    iLE = np.empty((len(hlist), 3), dtype="float")
    iRE = np.empty((len(hlist), 3), dtype="float")
    count = np.zeros(3, dtype="int64")
    pos = np.empty((npart, 3), dtype="float")
    for k in range(3):
        iLE[:, k] = DLE[k] + hdiv[k] * hpos[:, k]
        iRE[:, k] = iLE[:, k] + hdiv[k]
        iLE[hpos[:, k] != 0, k] -= buff * hdiv[k]
        iRE[hpos[:, k] != (dim_hilbert - 1), k] += buff * hdiv[k]
        gpos = np.clip(
            np.random.normal(DLE[k] + DW[k] / 2.0, DW[k] / 10.0, npart), DLE[k], DRE[k]
        )
        for ipos in gpos:
            for i in range(len(hlist)):
                if iLE[i, k] <= ipos < iRE[i, k]:
                    pos[count[k], k] = ipos
                    count[k] += 1
                    break
    return pos[: count.min(), :]


def fake_decomp_hilbert_uniform(
    npart, nfiles, ifile, DLE, DRE, buff=0.0, order=6, verbose=False
):
    np.random.seed(0x4D3D3D3 + ifile)
    DW = DRE - DLE
    dim_hilbert = 1 << order
    nH = dim_hilbert**3
    if nH < nfiles:
        raise Exception("Fewer hilbert cells than files.")
    nHPF = nH / nfiles
    rHPF = nH % nfiles
    nPH = npart / nH
    nRH = npart % nH
    hind = np.arange(nH, dtype="int64")
    hpos = get_hilbert_points(order, hind)
    hdiv = DW / dim_hilbert
    if ifile == 0:
        hlist = range(0, nHPF + rHPF)
        nptot = nPH * len(hlist) + nRH
    else:
        hlist = range(ifile * nHPF + rHPF, (ifile + 1) * nHPF + rHPF)
        nptot = nPH * len(hlist)
    pos = np.empty((nptot, 3), dtype="float")
    pc = 0
    for i in hlist:
        iLE = DLE + hdiv * hpos[i, :]
        iRE = iLE + hdiv
        for k in range(3):  # Don't add buffer past domain bounds
            if hpos[i, k] != 0:
                iLE[k] -= buff * hdiv[k]
            if hpos[i, k] != (dim_hilbert - 1):
                iRE[k] += buff * hdiv[k]
        inp = nPH
        if (ifile == 0) and (i == 0):
            inp += nRH
        for k in range(3):
            pos[pc : (pc + inp), k] = np.random.uniform(iLE[k], iRE[k], inp)
        pc += inp
    return pos


def fake_decomp_morton(
    npart, nfiles, ifile, DLE, DRE, buff=0.0, order=6, verbose=False
):
    np.random.seed(0x4D3D3D3 + ifile)
    DW = DRE - DLE
    dim_morton = 1 << order
    nH = dim_morton**3
    if nH < nfiles:
        raise Exception("Fewer morton cells than files.")
    nHPF = nH / nfiles
    rHPF = nH % nfiles
    nPH = npart / nH
    nRH = npart % nH
    hind = np.arange(nH, dtype="uint64")
    hpos = get_morton_points(hind)
    hdiv = DW / dim_morton
    if ifile == 0:
        hlist = range(0, nHPF + rHPF)
        nptot = nPH * len(hlist) + nRH
    else:
        hlist = range(ifile * nHPF + rHPF, (ifile + 1) * nHPF + rHPF)
        nptot = nPH * len(hlist)
    pos = np.empty((nptot, 3), dtype="float")
    pc = 0
    for i in hlist:
        iLE = DLE + hdiv * hpos[i, :]
        iRE = iLE + hdiv
        for k in range(3):  # Don't add buffer past domain bounds
            if hpos[i, k] != 0:
                iLE[k] -= buff * hdiv[k]
            if hpos[i, k] != (dim_morton - 1):
                iRE[k] += buff * hdiv[k]
        inp = nPH
        if (ifile == 0) and (i == 0):
            inp += nRH
        for k in range(3):
            pos[pc : (pc + inp), k] = np.random.uniform(iLE[k], iRE[k], inp)
        pc += inp
    return pos


def fake_decomp_grid(npart, nfiles, ifile, DLE, DRE, buff=0.0, verbose=False):
    # TODO: handle 'remainder' particles
    np.random.seed(0x4D3D3D3 + ifile)
    DW = DRE - DLE
    nYZ = int(np.sqrt(npart / nfiles))
    div = DW / nYZ
    Y, Z = np.mgrid[
        DLE[1] + 0.1 * div[1] : DRE[1] - 0.1 * div[1] : nYZ * 1j,
        DLE[2] + 0.1 * div[2] : DRE[2] - 0.1 * div[2] : nYZ * 1j,
    ]
    X = 0.5 * div[0] * np.ones(Y.shape, dtype="float64") + div[0] * ifile
    pos = np.array([X.ravel(), Y.ravel(), Z.ravel()], dtype="float64").transpose()
    return pos


def yield_fake_decomp(decomp, npart, nfiles, DLE, DRE, **kws):
    hsml = None
    for ifile in range(nfiles):
        yield fake_decomp(decomp, npart, nfiles, ifile, DLE, DRE, **kws), hsml


def fake_decomp(
    decomp, npart, nfiles, ifile, DLE, DRE, distrib="uniform", fname=None, **kws
):
    import pickle

    if fname is None and distrib == "gaussian":
        fname = f"{decomp}6_{distrib}_np{npart}_nf{nfiles}_file{ifile}"
    if fname is not None and os.path.isfile(fname):
        fd = open(fname, "rb")
        pos = pickle.load(fd)
        fd.close()
        return pos
    if decomp.startswith("zoom_"):
        zoom_factor = 5
        decomp_zoom = decomp.split("zoom_")[-1]
        zoom_npart = npart / 2
        zoom_rem = npart % 2
        pos1 = fake_decomp(
            decomp_zoom,
            zoom_npart + zoom_rem,
            nfiles,
            ifile,
            DLE,
            DRE,
            distrib=distrib,
            **kws,
        )
        DLE_zoom = DLE + 0.5 * DW * (1.0 - 1.0 / float(zoom_factor))
        DRE_zoom = DLE_zoom + DW / zoom_factor
        pos2 = fake_decomp(
            decomp_zoom,
            zoom_npart,
            nfiles,
            ifile,
            DLE_zoom,
            DRE_zoom,
            distrib=distrib,
            **kws,
        )
        pos = np.concatenate((pos1, pos2), axis=0)
    elif "_" in decomp:
        decomp_list = decomp.split("_")
        decomp_np = npart / len(decomp_list)
        decomp_nr = npart % len(decomp_list)
        pos = np.empty((0, 3), dtype="float")
        for i, idecomp in enumerate(decomp_list):
            inp = decomp_np
            if i == 0:
                inp += decomp_nr
            ipos = fake_decomp(
                idecomp, inp, nfiles, ifile, DLE, DRE, distrib=distrib, **kws
            )
            pos = np.concatenate((pos, ipos), axis=0)
    # A perfect grid, no overlap between files
    elif decomp == "grid":
        pos = fake_decomp_grid(npart, nfiles, ifile, DLE, DRE, **kws)
    # Completely random data set
    elif decomp == "random":
        if distrib == "uniform":
            pos = fake_decomp_random(npart, nfiles, ifile, DLE, DRE, **kws)
        else:
            raise ValueError(
                f"Unsupported value {distrib} for input parameter 'distrib'"
            )
    # Each file contains a slab (part of x domain, all of y/z domain)
    elif decomp == "sliced":
        if distrib == "uniform":
            pos = fake_decomp_sliced(npart, nfiles, ifile, DLE, DRE, **kws)
        else:
            raise ValueError(
                f"Unsupported value {distrib} for input parameter 'distrib'"
            )
    # Particles are assigned to files based on their location on a
    # Peano-Hilbert curve of order 6
    elif decomp.startswith("hilbert"):
        if decomp == "hilbert":
            kws["order"] = 6
        else:
            kws["order"] = int(decomp.split("hilbert")[-1])
        if distrib == "uniform":
            pos = fake_decomp_hilbert_uniform(npart, nfiles, ifile, DLE, DRE, **kws)
        elif distrib == "gaussian":
            makeall_decomp_hilbert_gaussian(
                npart, nfiles, DLE, DRE, fname_base=fname.split("file")[0], **kws
            )
            pos = fake_decomp(
                decomp,
                npart,
                nfiles,
                ifile,
                DLE,
                DRE,
                distrib=distrib,
                fname=fname,
                **kws,
            )
        else:
            raise ValueError(
                f"Unsupported value {distrib} for input parameter 'distrib'"
            )
    # Particles are assigned to files based on their location on a
    # Morton ordered Z-curve of order 6
    elif decomp.startswith("morton"):
        if decomp == "morton":
            kws["order"] = 6
        else:
            kws["order"] = int(decomp.split("morton")[-1])
        if distrib == "uniform":
            pos = fake_decomp_morton(npart, nfiles, ifile, DLE, DRE, **kws)
        else:
            raise ValueError(
                f"Unsupported value {distrib} for input parameter 'distrib'"
            )
    else:
        raise ValueError(f"Unsupported value {decomp} for input parameter 'decomp'")
    # Save
    if fname is not None:
        fd = open(fname, "wb")
        pickle.dump(pos, fd)
        fd.close()
    return pos



"""
Script to generate report of failed answer tests or to generate golden answers
on cloud platforms like Travis

"""

import argparse
import base64
import collections
import datetime
import logging
import os
import re
import shutil
import sys
import tempfile
import xml.etree.ElementTree as ET

import nose
import numpy
import requests

from yt.config import ytcfg
from yt.utilities.answer_testing.framework import AnswerTesting
from yt.utilities.command_line import FileStreamer

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("yt_report_failed_answers")
numpy.set_printoptions(threshold=5, edgeitems=1, precision=4)


def generate_failed_answers_html(failed_answers):
    """Generates html for the failed answer tests

    This function creates a html and embeds the images (actual, expected,
    difference) in it for the failed answers.

    Parameters
    ----------
    failed_answers : dict mapping string to dict
        the key is a string denoting the test name, the value is a
        dictionary that stores the actual, expected and difference plot
        file locations of the test.

    Returns
    -------
    string
        a html page

    """

    html_template = """
    <html><head>
    <style media="screen" type="text/css">
    img{{
      width:100%;
      max-width:800px;
    }}
    </style>
    <h1 style="text-align: center;">Failed Answer Tests</h1>
    <p>
      This report shows images of answer tests that failed when running
      the answer tests.
    </p>
    <p>
      <strong>Actual Image:</strong> plot generated while running the test<br/>
      <strong>Expected Image:</strong> golden answer image<br/>
      <strong>Difference Image:</strong> difference in the "actual"
      and "expected" image
    </p>
    <hr/>
    </head><body>
    <table>{rows}</table>
    </body></html>
    """

    row_template = """
    <tr>
    <td align="center">Actual</td>
    <td align="center">Expected</td>
    <td align="center">Difference</td>
    </tr>
    <tr>
    <td><img src="data:image/png;base64,{0}"></td>
    <td><img src="data:image/png;base64,{1}"></td>
    <td><img src="data:image/png;base64,{2}"></td>
    </tr>
    <tr><td align="center" colspan="3"><b>Test: {3}</b><hr/></td></tr>
    """

    rows = []

    for failed_test_file in failed_answers.values():
        for test_name, images in failed_test_file.items():
            encoded_images = {}
            for key in images:
                with open(images[key], "rb") as img:
                    img_data = base64.b64encode(img.read()).decode()
                    encoded_images[key] = img_data

            formatted_row = row_template.format(
                encoded_images["Actual"],
                encoded_images["Expected"],
                encoded_images["Difference"],
                test_name,
            )
            rows.append(formatted_row)

    html = html_template.format(rows="\n".join(rows))
    return html


def upload_to_curldrop(data, filename):
    """Uploads file to yt's curldrop server

    Uploads bytes `data` by the name `filename` to yt curldrop server.

    Parameters
    ----------
    data : bytes
        Content to be uploaded

    filename : string
        Name of file at curldrop's upload server

    Returns
    -------
    requests.models.Response
        Response returned by curldrop server

    """
    if "TRAVIS" in os.environ:
        job_num = os.environ["TRAVIS_JOB_NUMBER"]
        file_id = "Travis_Job_Num_" + job_num.replace(".", "_")
    else:
        file_id = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    filename = filename.format(file_id)

    base_url = ytcfg.get("yt", "curldrop_upload_url")
    upload_url = base_url + "/" + os.path.basename(filename)
    response = requests.put(upload_url, data=data)
    return response


def upload_failed_answers(failed_answers):
    """Uploads the result of failed answer tests

    Uploads a html page of the failed answer tests.

    Parameters
    ----------
    failed_answers : dict mapping string to dict
        the key is a string denoting the test name, the value is a
        dictionary that stores the actual, expected and difference plot
        file locations of the test.

    Returns
    -------
    requests.models.Response
        Response as returned by the upload service

    """
    html = generate_failed_answers_html(failed_answers)
    # convert html str to bytes
    html = html.encode()
    response = upload_to_curldrop(data=html, filename="failed_answers_{}.html")

    return response


def generate_answers(answer_dir, answers):
    """Generate golden answers

    Generates golden answers for the list of answers in ``answers`` and
    saves them at ``answer_dir``.

    Parameters
    ----------
    answer_dir : string
        directory location to save the generated answers

    answers : list of string
        Collection of missing answer tests specifying full name of the test.
        eg. ['yt.visualization.tests.test_line_plots:test_multi_line_plot']

    Returns
    -------
    bool
        True, if all the missing answers are successfully generated
        False, otherwise

    """
    status = True
    test_argv = [
        os.path.basename(__file__),
        "--with-answer-testing",
        "--nologcapture",
        "-s",
        "-d",
        "-v",
        "--local",
        f"--local-dir={answer_dir}",
        "--answer-store",
    ]

    for job in answers:
        log.info("\n Generating answers for %s", job)
        status &= nose.run(
            argv=test_argv + [job], addplugins=[AnswerTesting()], exit=False
        )
    return status


def upload_answers(answers):
    """Uploads answers not present in answer-store

    This function generates the answers for tests that are not present in
    answer store and uploads a zip file of the same.

    Parameters
    ----------
    answers : list of string
        Collection of missing answer tests specifying full name of the test.
        eg. ['yt.visualization.tests.test_line_plots:test_multi_line_plot']

    Returns
    -------
    requests.models.Response
        Response as returned by the upload service when answers are
        successfully uploaded

    None
        for the case when there was some error while generating the missing
        golden-answers

    """
    # Create temporary location to save new answers
    tmpdir = tempfile.mkdtemp()
    answer_dir = os.path.join(tmpdir, "answer-store")
    if not os.path.exists(answer_dir):
        os.mkdir(answer_dir)
    zip_file = os.path.join(tmpdir, "new-answers")

    status = generate_answers(answer_dir, answers)
    if status:
        zip_file = shutil.make_archive(zip_file, "zip", answer_dir)
        data = iter(FileStreamer(open(zip_file, "rb")))
        response = upload_to_curldrop(data=data, filename="new_answers_{}.zip")
        shutil.rmtree(tmpdir)
        return response
    return None


def extract_image_locations(error_string):
    """Regex based function to extract image file locations.

    Parameters
    ----------
    error_string : String
        The input string having file locations of 'Actual', 'Expected' and
        'Difference' plots. This string is generated by yt's answer-testing
        plugin, when the plot generated in the test does not match to its
        golden answer image.

    Returns
    -------
    dict
        If the `error_string` is successfully parsed to extract plot locations,
        then a dictionary with the keys 'Actual', 'Expected','Difference' and
        values having corresponding plot file locations is returned.
        eg. {'Actual': '/usr/tmp/tmp43la9b0w.png',
             'Expected': '/usr/tmp/tmpbpaqbgi3.png',
             'Difference': '/usr/tmp/tmp43la9b0w-failed-diff.png'}
    None
        When `error_string` does not conform to yt's answer-testing error
        message, which has the information for plot file locations on disk.

    """
    unknown_failure = False
    base_regex = r"\s*\n\s*(.*?.png)"
    img_regex = {
        "Actual": "Actual:" + base_regex,
        "Expected": "Expected:" + base_regex,
        "Difference": "Difference:" + base_regex,
    }
    img_path = {}
    for key in img_regex:
        result = re.search(img_regex[key], error_string, re.MULTILINE)
        if not result:
            unknown_failure = True
            break
        # store the locations of actual, expected and diff plot files
        img_path[key] = result.group(1)

    if not unknown_failure:
        return img_path
    return None


def parse_nose_xml(nose_xml):
    """Parse xml file generated by nosetests.

    Parse nose xml file to find following details:
        Failed tests: These could be due to difference in golden answer image
        and corresponding test plot.

        Missing tests: These errors occur when a corresponding golden answer
        image is not found.

    Parameters
    ----------
    nose_xml : string
        full path of xml file to be parsed

    Returns
    -------
    tuple : (failed_answers, missing_answers)

        failed_answers : list of tuples (string, dict)
        Collection of tuples where the first part is a string denoting the
        test name, the second part is a dictionary that stores the actual,
        expected and difference plot file locations of the test.
        eg. [('yt.visualization.tests.test_line_plots:test_line_plot',
                {'Actual': '/usr/tmp/tmp43la9b0w.png',
                'Expected': '/usr/tmp/tmpbpaqbgi3.png',
                'Difference': '/usr/tmp/tmp43la9b0w-failed-diff.png'}
            )]

        missing_answers : list of string
        Collection of missing answer tests specifying full name of the test.
        eg. ['yt.visualization.tests.test_line_plots:test_multi_line_plot']

    """
    missing_answers = set()
    failed_answers = collections.defaultdict(lambda: {})
    missing_errors = ["No such file or directory", "There is no old answer available"]
    tree = ET.parse(nose_xml)
    testsuite = tree.getroot()

    for testcase in testsuite:
        for error in testcase.iter("error"):
            handle_error(
                error, testcase, missing_errors, missing_answers, failed_answers
            )
        for error in testcase.iter("failure"):
            handle_error(
                error, testcase, missing_errors, missing_answers, failed_answers
            )
    return failed_answers, missing_answers


def handle_error(error, testcase, missing_errors, missing_answers, failed_answers):
    attribs = ["classname", "name"]
    test_name = ":".join(testcase.attrib[a] for a in attribs)
    message = error.attrib["message"]
    if (
        missing_errors[0] in error.attrib["message"]
        or missing_errors[1] in error.attrib["message"]
    ):
        missing_answers.add(test_name)
    elif "Items are not equal" in error.attrib["message"]:
        img_path = extract_image_locations(error.attrib["message"])
        if img_path:
            failed_answers[test_name][message] = img_path


if __name__ == "__main__":
    """Report failed answer tests of cloud platforms like Travis, Appveyor

    This script parses the nosetests xml file generated after answer tests are
    executed. If the test fail due to difference in actual and expected images,
    this function uploads a html page having all the plots which got failed
    (if executed with `-f` command line argument).
    In case, answer store does not has a golden answer and if executed with
    `-m` argument, it uploads missing answers zip file to yt's curldrop server.

    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f",
        "--upload-failed-tests",
        action="store_true",
        help="Upload a comparison report of failed answer tests"
        " to yt's curldrop server.",
    )
    parser.add_argument(
        "-m",
        "--upload-missing-answers",
        action="store_true",
        help="Upload tests' answers that are not found in answer-store.",
    )
    parser.add_argument(
        "--xunit-file",
        action="store",
        dest="nosetest_xml",
        required=True,
        help="Name of the nosetests xml file to parse for failed answer tests.",
    )
    args = parser.parse_args()

    # ANSI color codes
    COLOR_PURPLE = "\x1b[35;1m"
    COLOR_CYAN = "\x1b[36;1m"
    COLOR_RESET = "\x1b[0m"
    FLAG_EMOJI = " \U0001f6a9 "

    failed_answers = missing_answers = None
    if args.upload_failed_tests or args.upload_missing_answers:
        failed_answers, missing_answers = parse_nose_xml(args.nosetest_xml)

    if args.upload_failed_tests and failed_answers:
        response = upload_failed_answers(failed_answers)
        msg = ""
        if response.ok:
            msg += (
                "\n"
                + FLAG_EMOJI
                + COLOR_PURPLE
                + "Successfully uploaded failed answer test(s) result."
                " More details about the test failure can be found at the"
                " URL: "
                + response.text.split("\n")[1]
                + COLOR_RESET
                + FLAG_EMOJI
                + "\n"
            )
        response = upload_answers(failed_answers)
        if response is None:
            log.error("Failed to upload answers for failed tests !")
            sys.exit(1)
        if response.ok:
            msg += (
                FLAG_EMOJI
                + COLOR_CYAN
                + "Successfully uploaded answer(s) for failed test at URL: "
                + response.text.split("\n")[1]
                + " . Please commit these "
                "answers in the repository's answer-store." + COLOR_RESET + FLAG_EMOJI
            )
            log.info(msg)

    if args.upload_missing_answers and missing_answers:
        response = upload_answers(missing_answers)
        if response is None:
            log.error("Failed to upload missing answers !")
            sys.exit(1)
        if response.ok:
            msg = (
                FLAG_EMOJI
                + COLOR_CYAN
                + "Successfully uploaded missing answer(s) at URL: "
                + response.text.split("\n")[1]
                + " . Please commit these "
                "answers in the repository's answer-store." + COLOR_RESET + FLAG_EMOJI
            )
            log.info(msg)



from copy import deepcopy

import pytest

import yt
from yt.config import ytcfg
from yt.testing import requires_file

output_00080 = "output_00080/info_00080.txt"
ramses_new_format = "ramses_new_format/output_00002/info_00002.txt"

custom_hydro = [
    "my-Density",
    "my-x-velocity",
    "my-y-velocity",
    "my-z-velocity",
    "my-Pressure",
    "my-Metallicity",
]
custom_grav = [
    "my-x-acceleration",
    "my-y-acceleration",
    "my-z-acceleration",
]


@pytest.fixture()
def custom_ramses_fields_conf():
    old_config = deepcopy(ytcfg.config_root.as_dict())
    ytcfg.add_section("ramses-hydro")
    ytcfg.add_section("ramses-grav")
    ytcfg.set("ramses-hydro", "fields", custom_hydro)
    ytcfg.set("ramses-grav", "fields", custom_grav)
    yield
    ytcfg.remove_section("ramses-hydro")
    ytcfg.remove_section("ramses-grav")
    ytcfg.update(old_config)


@requires_file(output_00080)
def test_field_config_1(custom_ramses_fields_conf):
    ds = yt.load(output_00080)

    for f in custom_hydro:
        assert ("ramses", f) in ds.field_list
    for f in custom_grav:
        assert ("gravity", f) in ds.field_list


@requires_file(ramses_new_format)
def test_field_config_2(custom_ramses_fields_conf):
    ds = yt.load(ramses_new_format)

    for f in custom_hydro:
        assert ("ramses", f) in ds.field_list
    for f in custom_grav:
        assert ("gravity", f) in ds.field_list


@requires_file(output_00080)
@requires_file(ramses_new_format)
def test_warning_T2():
    ds1 = yt.load(output_00080)
    ds2 = yt.load(ramses_new_format)

    # Should not raise warnings
    ds1.r["gas", "temperature_over_mu"]
    ds2.r["gas", "temperature_over_mu"]

    # We cannot read the cooling tables of output_00080
    # so this should raise a warning
    with pytest.warns(
        RuntimeWarning,
        match=(
            "Trying to calculate temperature but the cooling tables couldn't be "
            r"found or read\. yt will return T/µ instead of T.*"
        ),
    ):
        ds1.r["gas", "temperature"]

    # But this one should not
    ds2.r["gas", "temperature"]



import base64
import os
from functools import wraps
from typing import TYPE_CHECKING, Any

import matplotlib
import numpy as np
from more_itertools.more import always_iterable, unzip

from yt._maintenance.ipython_compat import IS_IPYTHON
from yt._typing import FieldKey
from yt.data_objects.profiles import create_profile, sanitize_field_tuple_keys
from yt.data_objects.static_output import Dataset
from yt.frontends.ytdata.data_structures import YTProfileDataset
from yt.funcs import iter_fields, matplotlib_style_context
from yt.utilities.exceptions import YTNotInsideNotebook
from yt.visualization._commons import _get_units_label
from yt.visualization._handlers import ColorbarHandler, NormHandler
from yt.visualization.base_plot_types import ImagePlotMPL, PlotMPL

from ..data_objects.selection_objects.data_selection_objects import YTSelectionContainer
from ._commons import validate_image_name
from .plot_container import (
    BaseLinePlot,
    ImagePlotContainer,
    invalidate_plot,
    validate_plot,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from yt._typing import FieldKey


def invalidate_profile(f):
    @wraps(f)
    def newfunc(*args, **kwargs):
        rv = f(*args, **kwargs)
        args[0]._profile_valid = False
        return rv

    return newfunc


def sanitize_label(labels, nprofiles):
    labels = list(always_iterable(labels)) or [None]

    if len(labels) == 1:
        labels = labels * nprofiles

    if len(labels) != nprofiles:
        raise ValueError(
            f"Number of labels {len(labels)} must match number of profiles {nprofiles}"
        )

    invalid_data = [
        (label, type(label))
        for label in labels
        if label is not None and not isinstance(label, str)
    ]
    if invalid_data:
        invalid_labels, types = unzip(invalid_data)
        raise TypeError(
            "All labels must be None or a string, "
            f"received {invalid_labels} with type {types}"
        )

    return labels


def data_object_or_all_data(data_source):
    if isinstance(data_source, Dataset):
        data_source = data_source.all_data()

    if not isinstance(data_source, YTSelectionContainer):
        raise RuntimeError("data_source must be a yt selection data object")

    return data_source


class ProfilePlot(BaseLinePlot):
    r"""
    Create a 1d profile plot from a data source or from a list
    of profile objects.

    Given a data object (all_data, region, sphere, etc.), an x field,
    and a y field (or fields), this will create a one-dimensional profile
    of the average (or total) value of the y field in bins of the x field.

    This can be used to create profiles from given fields or to plot
    multiple profiles created from
    `yt.data_objects.profiles.create_profile`.

    Parameters
    ----------
    data_source : YTSelectionContainer Object
        The data object to be profiled, such as all_data, region, or
        sphere. If a dataset is passed in instead, an all_data data object
        is generated internally from the dataset.
    x_field : str
        The binning field for the profile.
    y_fields : str or list
        The field or fields to be profiled.
    weight_field : str
        The weight field for calculating weighted averages. If None,
        the profile values are the sum of the field values within the bin.
        Otherwise, the values are a weighted average.
        Default : ("gas", "mass")
    n_bins : int
        The number of bins in the profile.
        Default: 64.
    accumulation : bool
        If True, the profile values for a bin N are the cumulative sum of
        all the values from bin 0 to N.
        Default: False.
    fractional : If True the profile values are divided by the sum of all
        the profile data such that the profile represents a probability
        distribution function.
    label : str or list of strings
        If a string, the label to be put on the line plotted.  If a list,
        this should be a list of labels for each profile to be overplotted.
        Default: None.
    plot_spec : dict or list of dicts
        A dictionary or list of dictionaries containing plot keyword
        arguments.  For example, dict(color="red", linestyle=":").
        Default: None.
    x_log : bool
        Whether the x_axis should be plotted with a logarithmic
        scaling (True), or linear scaling (False).
        Default: True.
    y_log : dict or bool
        A dictionary containing field:boolean pairs, setting the logarithmic
        property for that field. May be overridden after instantiation using
        set_log
        A single boolean can be passed to signify all fields should use
        logarithmic (True) or linear scaling (False).
        Default: True.

    Examples
    --------

    This creates profiles of a single dataset.

    >>> import yt
    >>> ds = yt.load("enzo_tiny_cosmology/DD0046/DD0046")
    >>> ad = ds.all_data()
    >>> plot = yt.ProfilePlot(
    ...     ad,
    ...     ("gas", "density"),
    ...     [("gas", "temperature"), ("gas", "velocity_x")],
    ...     weight_field=("gas", "mass"),
    ...     plot_spec=dict(color="red", linestyle="--"),
    ... )
    >>> plot.save()

    This creates profiles from a time series object.

    >>> es = yt.load_simulation("AMRCosmology.enzo", "Enzo")
    >>> es.get_time_series()

    >>> profiles = []
    >>> labels = []
    >>> plot_specs = []
    >>> for ds in es[-4:]:
    ...     ad = ds.all_data()
    ...     profiles.append(
    ...         create_profile(
    ...             ad,
    ...             [("gas", "density")],
    ...             fields=[("gas", "temperature"), ("gas", "velocity_x")],
    ...         )
    ...     )
    ...     labels.append(ds.current_redshift)
    ...     plot_specs.append(dict(linestyle="--", alpha=0.7))

    >>> plot = yt.ProfilePlot.from_profiles(
    ...     profiles, labels=labels, plot_specs=plot_specs
    ... )
    >>> plot.save()

    Use set_line_property to change line properties of one or all profiles.

    """

    _default_figure_size = (10.0, 8.0)
    _default_font_size = 18.0

    x_log = None
    y_log = None
    x_title = None
    y_title = None
    _plot_valid = False

    def __init__(
        self,
        data_source,
        x_field,
        y_fields,
        weight_field=("gas", "mass"),
        n_bins=64,
        accumulation=False,
        fractional=False,
        label=None,
        plot_spec=None,
        x_log=True,
        y_log=True,
    ):
        data_source = data_object_or_all_data(data_source)
        y_fields = list(iter_fields(y_fields))
        logs = {x_field: bool(x_log)}
        if isinstance(y_log, bool):
            y_log = {y_field: y_log for y_field in y_fields}

        if isinstance(data_source.ds, YTProfileDataset):
            profiles = [data_source.ds.profile]
        else:
            profiles = [
                create_profile(
                    data_source,
                    [x_field],
                    n_bins=[n_bins],
                    fields=y_fields,
                    weight_field=weight_field,
                    accumulation=accumulation,
                    fractional=fractional,
                    logs=logs,
                )
            ]

        if plot_spec is None:
            plot_spec = [{} for p in profiles]
        if not isinstance(plot_spec, list):
            plot_spec = [plot_spec.copy() for p in profiles]

        ProfilePlot._initialize_instance(
            self, data_source, profiles, label, plot_spec, y_log
        )

    @classmethod
    def _initialize_instance(
        cls,
        obj,
        data_source,
        profiles,
        labels,
        plot_specs,
        y_log,
    ):
        obj._plot_title = {}
        obj._plot_text = {}
        obj._text_xpos = {}
        obj._text_ypos = {}
        obj._text_kwargs = {}

        super(ProfilePlot, obj).__init__(data_source)
        obj.profiles = list(always_iterable(profiles))
        obj.x_log = None
        obj.y_log = sanitize_field_tuple_keys(y_log, data_source) or {}
        obj.y_title = {}
        obj.x_title = None
        obj.label = sanitize_label(labels, len(obj.profiles))
        if plot_specs is None:
            plot_specs = [{} for p in obj.profiles]
        obj.plot_spec = plot_specs
        obj._xlim = (None, None)
        obj._setup_plots()
        obj._plot_valid = False  # see https://github.com/yt-project/yt/issues/4489
        return obj

    def _get_axrect(self):
        return (0.1, 0.1, 0.8, 0.8)

    @validate_plot
    def save(
        self,
        name: str | None = None,
        suffix: str | None = None,
        mpl_kwargs: dict[str, Any] | None = None,
    ):
        r"""
        Saves a 1d profile plot.

        Parameters
        ----------
        name : str, optional
            The output file keyword.
        suffix : string, optional
            Specify the image type by its suffix. If not specified, the output
            type will be inferred from the filename. Defaults to '.png'.
        mpl_kwargs : dict, optional
            A dict of keyword arguments to be passed to matplotlib.
        """
        if not self._plot_valid:
            self._setup_plots()

        # Mypy is hardly convinced that we have a `profiles` attribute
        # at this stage, so we're lasily going to deactivate it locally
        unique = set(self.plots.values())
        iters: Iterable[tuple[int | FieldKey, PlotMPL]]
        if len(unique) < len(self.plots):
            iters = enumerate(sorted(unique))
        else:
            iters = self.plots.items()

        if name is None:
            if len(self.profiles) == 1:  # type: ignore
                name = str(self.profiles[0].ds)  # type: ignore
            else:
                name = "Multi-data"

        name = validate_image_name(name, suffix)
        prefix, suffix = os.path.splitext(name)

        xfn = self.profiles[0].x_field  # type: ignore
        if isinstance(xfn, tuple):
            xfn = xfn[1]

        names = []
        for uid, plot in iters:
            if isinstance(uid, tuple):
                uid = uid[1]  # type: ignore
            uid_name = f"{prefix}_1d-Profile_{xfn}_{uid}{suffix}"
            names.append(uid_name)
            with matplotlib_style_context():
                plot.save(uid_name, mpl_kwargs=mpl_kwargs)
        return names

    @validate_plot
    def show(self):
        r"""This will send any existing plots to the IPython notebook.

        If yt is being run from within an IPython session, and it is able to
        determine this, this function will send any existing plots to the
        notebook for display.

        If yt can't determine if it's inside an IPython session, it will raise
        YTNotInsideNotebook.

        Examples
        --------

        >>> import yt
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> pp = ProfilePlot(ds.all_data(), ("gas", "density"), ("gas", "temperature"))
        >>> pp.show()

        """
        if IS_IPYTHON:
            from IPython.display import display

            display(self)
        else:
            raise YTNotInsideNotebook

    @validate_plot
    def _repr_html_(self):
        """Return an html representation of the plot object. Will display as a
        png for each WindowPlotMPL instance in self.plots"""
        ret = ""
        unique = set(self.plots.values())
        if len(unique) < len(self.plots):
            iters = sorted(unique)
        else:
            iters = self.plots.values()
        for plot in iters:
            with matplotlib_style_context():
                img = plot._repr_png_()
            img = base64.b64encode(img).decode()
            ret += (
                r'<img style="max-width:100%;max-height:100%;" '
                rf'src="data:image/png;base64,{img}"><br>'
            )
        return ret

    def _setup_plots(self):
        if self._plot_valid:
            return
        for f, p in self.plots.items():
            p.axes.cla()
            if f in self._plot_text:
                p.axes.text(
                    self._text_xpos[f],
                    self._text_ypos[f],
                    self._plot_text[f],
                    fontproperties=self._font_properties,
                    **self._text_kwargs[f],
                )
        self._set_font_properties()

        for i, profile in enumerate(self.profiles):
            for field, field_data in profile.items():
                plot = self._get_plot_instance(field)
                plot.axes.plot(
                    np.array(profile.x),
                    np.array(field_data),
                    label=self.label[i],
                    **self.plot_spec[i],
                )

        for profile in self.profiles:
            for fname in profile.keys():
                axes = self.plots[fname].axes
                xscale, yscale = self._get_field_log(fname, profile)
                xtitle, ytitle = self._get_field_title(fname, profile)

                axes.set_xscale(xscale)
                axes.set_yscale(yscale)

                axes.set_ylabel(ytitle)
                axes.set_xlabel(xtitle)

                pnh = self.plots[fname].norm_handler

                axes.set_ylim(pnh.vmin, pnh.vmax)
                axes.set_xlim(*self._xlim)

                if fname in self._plot_title:
                    axes.set_title(self._plot_title[fname])

                if any(self.label):
                    axes.legend(loc="best")
        self._set_font_properties()
        self._plot_valid = True

    @classmethod
    def from_profiles(cls, profiles, labels=None, plot_specs=None, y_log=None):
        r"""
        Instantiate a ProfilePlot object from a list of profiles
        created with :func:`~yt.data_objects.profiles.create_profile`.

        Parameters
        ----------
        profiles : a profile or list of profiles
            A single profile or list of profile objects created with
            :func:`~yt.data_objects.profiles.create_profile`.
        labels : list of strings
            A list of labels for each profile to be overplotted.
            Default: None.
        plot_specs : list of dicts
            A list of dictionaries containing plot keyword
            arguments.  For example, [dict(color="red", linestyle=":")].
            Default: None.

        Examples
        --------

        >>> from yt import load_simulation
        >>> es = load_simulation("AMRCosmology.enzo", "Enzo")
        >>> es.get_time_series()

        >>> profiles = []
        >>> labels = []
        >>> plot_specs = []
        >>> for ds in es[-4:]:
        ...     ad = ds.all_data()
        ...     profiles.append(
        ...         create_profile(
        ...             ad,
        ...             [("gas", "density")],
        ...             fields=[("gas", "temperature"), ("gas", "velocity_x")],
        ...         )
        ...     )
        ...     labels.append(ds.current_redshift)
        ...     plot_specs.append(dict(linestyle="--", alpha=0.7))
        >>> plot = ProfilePlot.from_profiles(
        ...     profiles, labels=labels, plot_specs=plot_specs
        ... )
        >>> plot.save()

        """
        if labels is not None and len(profiles) != len(labels):
            raise RuntimeError("Profiles list and labels list must be the same size.")
        if plot_specs is not None and len(plot_specs) != len(profiles):
            raise RuntimeError(
                "Profiles list and plot_specs list must be the same size."
            )
        obj = cls.__new__(cls)
        profiles = list(always_iterable(profiles))
        return cls._initialize_instance(
            obj, profiles[0].data_source, profiles, labels, plot_specs, y_log
        )

    @invalidate_plot
    def set_line_property(self, property, value, index=None):
        r"""
        Set properties for one or all lines to be plotted.

        Parameters
        ----------
        property : str
            The line property to be set.
        value : str, int, float
            The value to set for the line property.
        index : int
            The index of the profile in the list of profiles to be
            changed.  If None, change all plotted lines.
            Default : None.

        Examples
        --------

        Change all the lines in a plot
        plot.set_line_property("linestyle", "-")

        Change a single line.
        plot.set_line_property("linewidth", 4, index=0)

        """
        if index is None:
            specs = self.plot_spec
        else:
            specs = [self.plot_spec[index]]
        for spec in specs:
            spec[property] = value
        return self

    @invalidate_plot
    def set_log(self, field, log):
        """set a field to log or linear.

        Parameters
        ----------
        field : string
            the field to set a transform
        log : boolean
            Log on/off.
        """
        if field == "all":
            self.x_log = log
            for field in list(self.profiles[0].field_data.keys()):
                self.y_log[field] = log
        else:
            (field,) = self.profiles[0].data_source._determine_fields([field])
            if field == self.profiles[0].x_field:
                self.x_log = log
            elif field in self.profiles[0].field_data:
                self.y_log[field] = log
            else:
                raise KeyError(f"Field {field} not in profile plot!")
        return self

    @invalidate_plot
    def set_ylabel(self, field, label):
        """Sets a new ylabel for the specified fields

        Parameters
        ----------
        field : string
           The name of the field that is to be changed.

        label : string
           The label to be placed on the y-axis
        """
        if field == "all":
            for field in self.profiles[0].field_data:
                self.y_title[field] = label
        else:
            (field,) = self.profiles[0].data_source._determine_fields([field])
            if field in self.profiles[0].field_data:
                self.y_title[field] = label
            else:
                raise KeyError(f"Field {field} not in profile plot!")

        return self

    @invalidate_plot
    def set_xlabel(self, label):
        """Sets a new xlabel for all profiles

        Parameters
        ----------
        label : string
           The label to be placed on the x-axis
        """
        self.x_title = label

        return self

    @invalidate_plot
    def set_unit(self, field, unit):
        """Sets a new unit for the requested field

        Parameters
        ----------
        field : string
           The name of the field that is to be changed.

        unit : string or Unit object
           The name of the new unit.
        """
        fd = self.profiles[0].data_source._determine_fields(field)[0]
        for profile in self.profiles:
            if fd == profile.x_field:
                profile.set_x_unit(unit)
            elif fd[1] in self.profiles[0].field_map:
                profile.set_field_unit(field, unit)
            else:
                raise KeyError(f"Field {field} not in profile plot!")
        return self

    @invalidate_plot
    def set_xlim(self, xmin=None, xmax=None):
        """Sets the limits of the bin field

        Parameters
        ----------

        xmin : float or None
          The new x minimum.  Defaults to None, which leaves the xmin
          unchanged.

        xmax : float or None
          The new x maximum.  Defaults to None, which leaves the xmax
          unchanged.

        Examples
        --------

        >>> import yt
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> pp = yt.ProfilePlot(
        ...     ds.all_data(), ("gas", "density"), ("gas", "temperature")
        ... )
        >>> pp.set_xlim(1e-29, 1e-24)
        >>> pp.save()

        """
        self._xlim = (xmin, xmax)
        for i, p in enumerate(self.profiles):
            if xmin is None:
                xmi = p.x_bins.min()
            else:
                xmi = xmin
            if xmax is None:
                xma = p.x_bins.max()
            else:
                xma = xmax
            extrema = {p.x_field: ((xmi, str(p.x.units)), (xma, str(p.x.units)))}
            units = {p.x_field: str(p.x.units)}
            if self.x_log is None:
                logs = None
            else:
                logs = {p.x_field: self.x_log}
            for field in p.field_map.values():
                units[field] = str(p.field_data[field].units)
            self.profiles[i] = create_profile(
                p.data_source,
                p.x_field,
                n_bins=len(p.x_bins) - 1,
                fields=list(p.field_map.values()),
                weight_field=p.weight_field,
                accumulation=p.accumulation,
                fractional=p.fractional,
                logs=logs,
                extrema=extrema,
                units=units,
            )
        return self

    @invalidate_plot
    def set_ylim(self, field, ymin=None, ymax=None):
        """Sets the plot limits for the specified field we are binning.

        Parameters
        ----------

        field : string or field tuple

        The field that we want to adjust the plot limits for.

        ymin : float or None
          The new y minimum.  Defaults to None, which leaves the ymin
          unchanged.

        ymax : float or None
          The new y maximum.  Defaults to None, which leaves the ymax
          unchanged.

        Examples
        --------

        >>> import yt
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> pp = yt.ProfilePlot(
        ...     ds.all_data(),
        ...     ("gas", "density"),
        ...     [("gas", "temperature"), ("gas", "velocity_x")],
        ... )
        >>> pp.set_ylim(("gas", "temperature"), 1e4, 1e6)
        >>> pp.save()

        """
        fields = list(self.plots.keys()) if field == "all" else field
        for profile in self.profiles:
            for field in profile.data_source._determine_fields(fields):
                if field in profile.field_map:
                    field = profile.field_map[field]
                pnh = self.plots[field].norm_handler
                pnh.vmin = ymin
                pnh.vmax = ymax
                # Continue on to the next profile.
                break
        return self

    def _set_font_properties(self):
        for f in self.plots:
            self.plots[f]._set_font_properties(self._font_properties, self._font_color)

    def _get_field_log(self, field_y, profile):
        yfi = profile.field_info[field_y]
        if self.x_log is None:
            x_log = profile.x_log
        else:
            x_log = self.x_log
        y_log = self.y_log.get(field_y, yfi.take_log)
        scales = {True: "log", False: "linear"}
        return scales[x_log], scales[y_log]

    def _get_field_label(self, field, field_info, field_unit, fractional=False):
        field_unit = field_unit.latex_representation()
        field_name = field_info.display_name
        if isinstance(field, tuple):
            field = field[1]
        if field_name is None:
            field_name = field_info.get_latex_display_name()
        elif field_name.find("$") == -1:
            field_name = field_name.replace(" ", r"\ ")
            field_name = r"$\rm{" + field_name + r"}$"
        if fractional:
            label = field_name + r"$\rm{\ Probability\ Density}$"
        elif field_unit is None or field_unit == "":
            label = field_name
        else:
            label = field_name + _get_units_label(field_unit)
        return label

    def _get_field_title(self, field_y, profile):
        field_x = profile.x_field
        xfi = profile.field_info[field_x]
        yfi = profile.field_info[field_y]
        x_unit = profile.x.units
        y_unit = profile.field_units[field_y]
        fractional = profile.fractional
        x_title = self.x_title or self._get_field_label(field_x, xfi, x_unit)
        y_title = self.y_title.get(field_y, None) or self._get_field_label(
            field_y, yfi, y_unit, fractional
        )

        return (x_title, y_title)

    @invalidate_plot
    def annotate_title(self, title, field="all"):
        r"""Set a title for the plot.

        Parameters
        ----------
        title : str
          The title to add.
        field : str or list of str
          The field name for which title needs to be set.

        Examples
        --------
        >>> # To set title for all the fields:
        >>> plot.annotate_title("This is a Profile Plot")

        >>> # To set title for specific fields:
        >>> plot.annotate_title("Profile Plot for Temperature", ("gas", "temperature"))

        >>> # Setting same plot title for both the given fields
        >>> plot.annotate_title(
        ...     "Profile Plot: Temperature-Dark Matter Density",
        ...     [("gas", "temperature"), ("deposit", "dark_matter_density")],
        ... )

        """
        fields = list(self.plots.keys()) if field == "all" else field
        for profile in self.profiles:
            for field in profile.data_source._determine_fields(fields):
                if field in profile.field_map:
                    field = profile.field_map[field]
                self._plot_title[field] = title
        return self

    @invalidate_plot
    def annotate_text(self, xpos=0.0, ypos=0.0, text=None, field="all", **text_kwargs):
        r"""Allow the user to insert text onto the plot

        The x-position and y-position must be given as well as the text string.
        Add *text* to plot at location *xpos*, *ypos* in plot coordinates for
        the given fields or by default for all fields.
        (see example below).

        Parameters
        ----------
        xpos : float
          Position on plot in x-coordinates.
        ypos : float
          Position on plot in y-coordinates.
        text : str
          The text to insert onto the plot.
        field : str or tuple
          The name of the field to add text to.
        **text_kwargs : dict
          Extra keyword arguments will be passed to matplotlib text instance

        >>> import yt
        >>> from yt.units import kpc
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> my_galaxy = ds.disk(ds.domain_center, [0.0, 0.0, 1.0], 10 * kpc, 3 * kpc)
        >>> plot = yt.ProfilePlot(
        ...     my_galaxy, ("gas", "density"), [("gas", "temperature")]
        ... )

        >>> # Annotate text for all the fields
        >>> plot.annotate_text(1e-26, 1e5, "This is annotated text in the plot area.")
        >>> plot.save()

        >>> # Annotate text for a given field
        >>> plot.annotate_text(1e-26, 1e5, "Annotated text", ("gas", "temperature"))
        >>> plot.save()

        >>> # Annotate text for multiple fields
        >>> fields = [("gas", "temperature"), ("gas", "density")]
        >>> plot.annotate_text(1e-26, 1e5, "Annotated text", fields)
        >>> plot.save()

        """
        fields = list(self.plots.keys()) if field == "all" else field
        for profile in self.profiles:
            for field in profile.data_source._determine_fields(fields):
                if field in profile.field_map:
                    field = profile.field_map[field]
                self._plot_text[field] = text
                self._text_xpos[field] = xpos
                self._text_ypos[field] = ypos
                self._text_kwargs[field] = text_kwargs
        return self


class PhasePlot(ImagePlotContainer):
    r"""
    Create a 2d profile (phase) plot from a data source or from
    profile object created with
    `yt.data_objects.profiles.create_profile`.

    Given a data object (all_data, region, sphere, etc.), an x field,
    y field, and z field (or fields), this will create a two-dimensional
    profile of the average (or total) value of the z field in bins of the
    x and y fields.

    Parameters
    ----------
    data_source : YTSelectionContainer Object
        The data object to be profiled, such as all_data, region, or
        sphere. If a dataset is passed in instead, an all_data data object
        is generated internally from the dataset.
    x_field : str
        The x binning field for the profile.
    y_field : str
        The y binning field for the profile.
    z_fields : str or list
        The field or fields to be profiled.
    weight_field : str
        The weight field for calculating weighted averages.  If None,
        the profile values are the sum of the field values within the bin.
        Otherwise, the values are a weighted average.
        Default : ("gas", "mass")
    x_bins : int
        The number of bins in x field for the profile.
        Default: 128.
    y_bins : int
        The number of bins in y field for the profile.
        Default: 128.
    accumulation : bool or list of bools
        If True, the profile values for a bin n are the cumulative sum of
        all the values from bin 0 to n.  If -True, the sum is reversed so
        that the value for bin n is the cumulative sum from bin N (total bins)
        to n.  A list of values can be given to control the summation in each
        dimension independently.
        Default: False.
    fractional : If True the profile values are divided by the sum of all
        the profile data such that the profile represents a probability
        distribution function.
    fontsize : int
        Font size for all text in the plot.
        Default: 18.
    figure_size : int
        Size in inches of the image.
        Default: 8 (8x8)
    shading : str
        This argument is directly passed down to matplotlib.axes.Axes.pcolormesh
        see
        https://matplotlib.org/3.3.1/gallery/images_contours_and_fields/pcolormesh_grids.html#sphx-glr-gallery-images-contours-and-fields-pcolormesh-grids-py  # noqa
        Default: 'nearest'

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("enzo_tiny_cosmology/DD0046/DD0046")
    >>> ad = ds.all_data()
    >>> plot = yt.PhasePlot(
    ...     ad,
    ...     ("gas", "density"),
    ...     ("gas", "temperature"),
    ...     [("gas", "mass")],
    ...     weight_field=None,
    ... )
    >>> plot.save()

    >>> # Change plot properties.
    >>> plot.set_cmap(("gas", "mass"), "jet")
    >>> plot.set_zlim(("gas", "mass"), 1e8, 1e13)
    >>> plot.annotate_title("This is a phase plot")

    """

    x_log = None
    y_log = None
    plot_title = None
    _plot_valid = False
    _profile_valid = False
    _plot_type = "Phase"
    _xlim = (None, None)
    _ylim = (None, None)

    def __init__(
        self,
        data_source,
        x_field,
        y_field,
        z_fields,
        weight_field=("gas", "mass"),
        x_bins=128,
        y_bins=128,
        accumulation=False,
        fractional=False,
        fontsize=18,
        figure_size=8.0,
        shading="nearest",
    ):
        data_source = data_object_or_all_data(data_source)

        if isinstance(z_fields, tuple):
            z_fields = [z_fields]
        z_fields = list(always_iterable(z_fields))

        if isinstance(data_source.ds, YTProfileDataset):
            profile = data_source.ds.profile
        else:
            profile = create_profile(
                data_source,
                [x_field, y_field],
                z_fields,
                n_bins=[x_bins, y_bins],
                weight_field=weight_field,
                accumulation=accumulation,
                fractional=fractional,
            )

        type(self)._initialize_instance(
            self, data_source, profile, fontsize, figure_size, shading
        )

    @classmethod
    def _initialize_instance(
        cls, obj, data_source, profile, fontsize, figure_size, shading
    ):
        obj.plot_title = {}
        obj.z_log = {}
        obj.z_title = {}
        obj._initfinished = False
        obj.x_log = None
        obj.y_log = None
        obj._plot_text = {}
        obj._text_xpos = {}
        obj._text_ypos = {}
        obj._text_kwargs = {}
        obj._profile = profile
        obj._shading = shading
        obj._profile_valid = True
        obj._xlim = (None, None)
        obj._ylim = (None, None)
        super(PhasePlot, obj).__init__(data_source, figure_size, fontsize)
        obj._setup_plots()
        obj._plot_valid = False  # see https://github.com/yt-project/yt/issues/4489
        obj._initfinished = True
        return obj

    def _get_field_title(self, field_z, profile):
        field_x = profile.x_field
        field_y = profile.y_field
        xfi = profile.field_info[field_x]
        yfi = profile.field_info[field_y]
        zfi = profile.field_info[field_z]
        x_unit = profile.x.units
        y_unit = profile.y.units
        z_unit = profile.field_units[field_z]
        fractional = profile.fractional
        x_label, y_label, z_label = self._get_axes_labels(field_z)
        x_title = x_label or self._get_field_label(field_x, xfi, x_unit)
        y_title = y_label or self._get_field_label(field_y, yfi, y_unit)
        z_title = z_label or self._get_field_label(field_z, zfi, z_unit, fractional)
        return (x_title, y_title, z_title)

    def _get_field_label(self, field, field_info, field_unit, fractional=False):
        field_unit = field_unit.latex_representation()
        field_name = field_info.display_name
        if isinstance(field, tuple):
            field = field[1]
        if field_name is None:
            field_name = field_info.get_latex_display_name()
        elif field_name.find("$") == -1:
            field_name = field_name.replace(" ", r"\ ")
            field_name = r"$\rm{" + field_name + r"}$"
        if fractional:
            label = field_name + r"$\rm{\ Probability\ Density}$"
        elif field_unit is None or field_unit == "":
            label = field_name
        else:
            label = field_name + _get_units_label(field_unit)
        return label

    def _get_field_log(self, field_z, profile):
        zfi = profile.field_info[field_z]
        if self.x_log is None:
            x_log = profile.x_log
        else:
            x_log = self.x_log
        if self.y_log is None:
            y_log = profile.y_log
        else:
            y_log = self.y_log
        if field_z in self.z_log:
            z_log = self.z_log[field_z]
        else:
            z_log = zfi.take_log
        scales = {True: "log", False: "linear"}
        return scales[x_log], scales[y_log], scales[z_log]

    @property
    def profile(self):
        if not self._profile_valid:
            self._recreate_profile()
        return self._profile

    @property
    def fields(self):
        return list(self.plots.keys())

    def _setup_plots(self):
        if self._plot_valid:
            return
        for f, data in self.profile.items():
            if f in self.plots:
                pnh = self.plots[f].norm_handler
                cbh = self.plots[f].colorbar_handler
                draw_axes = self.plots[f]._draw_axes
                if self.plots[f].figure is not None:
                    fig = self.plots[f].figure
                    axes = self.plots[f].axes
                    cax = self.plots[f].cax
                else:
                    fig = None
                    axes = None
                    cax = None
            else:
                pnh, cbh = self._get_default_handlers(
                    field=f, default_display_units=self.profile[f].units
                )
                fig = None
                axes = None
                cax = None
                draw_axes = True

            x_scale, y_scale, z_scale = self._get_field_log(f, self.profile)
            x_title, y_title, z_title = self._get_field_title(f, self.profile)

            font_size = self._font_properties.get_size()
            f = self.profile.data_source._determine_fields(f)[0]

            # if this is a Particle Phase Plot AND if we using a single color,
            # override the colorbar here.
            splat_color = getattr(self, "splat_color", None)
            if splat_color is not None:
                cbh.cmap = matplotlib.colors.ListedColormap(splat_color, "dummy")

            masked_data = data.copy()
            masked_data[~self.profile.used] = np.nan
            self.plots[f] = PhasePlotMPL(
                self.profile.x,
                self.profile.y,
                masked_data,
                x_scale,
                y_scale,
                self.figure_size,
                font_size,
                fig,
                axes,
                cax,
                shading=self._shading,
                norm_handler=pnh,
                colorbar_handler=cbh,
            )

            self.plots[f]._toggle_axes(draw_axes)
            self.plots[f]._toggle_colorbar(cbh.draw_cbar)

            self.plots[f].axes.xaxis.set_label_text(x_title)
            self.plots[f].axes.yaxis.set_label_text(y_title)
            self.plots[f].cax.yaxis.set_label_text(z_title)

            self.plots[f].axes.set_xlim(self._xlim)
            self.plots[f].axes.set_ylim(self._ylim)

            if f in self._plot_text:
                self.plots[f].axes.text(
                    self._text_xpos[f],
                    self._text_ypos[f],
                    self._plot_text[f],
                    fontproperties=self._font_properties,
                    **self._text_kwargs[f],
                )

            if f in self.plot_title:
                self.plots[f].axes.set_title(self.plot_title[f])

            # x-y axes minorticks
            if f not in self._minorticks:
                self._minorticks[f] = True
            if self._minorticks[f]:
                self.plots[f].axes.minorticks_on()
            else:
                self.plots[f].axes.minorticks_off()

        self._set_font_properties()

        # if this is a particle plot with one color only, hide the cbar here
        if hasattr(self, "use_cbar") and not self.use_cbar:
            self.plots[f].hide_colorbar()

        self._plot_valid = True

    @classmethod
    def from_profile(cls, profile, fontsize=18, figure_size=8.0, shading="nearest"):
        r"""
        Instantiate a PhasePlot object from a profile object created
        with :func:`~yt.data_objects.profiles.create_profile`.

        Parameters
        ----------
        profile : An instance of :class:`~yt.data_objects.profiles.ProfileND`
             A single profile object.
        fontsize : float
             The fontsize to use, in points.
        figure_size : float
             The figure size to use, in inches.
        shading : str
            This argument is directly passed down to matplotlib.axes.Axes.pcolormesh
            see
            https://matplotlib.org/3.3.1/gallery/images_contours_and_fields/pcolormesh_grids.html#sphx-glr-gallery-images-contours-and-fields-pcolormesh-grids-py  # noqa
            Default: 'nearest'

        Examples
        --------

        >>> import yt
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> extrema = {
        ...     ("gas", "density"): (1e-31, 1e-24),
        ...     ("gas", "temperature"): (1e1, 1e8),
        ...     ("gas", "mass"): (1e-6, 1e-1),
        ... }
        >>> profile = yt.create_profile(
        ...     ds.all_data(),
        ...     [("gas", "density"), ("gas", "temperature")],
        ...     fields=[("gas", "mass")],
        ...     extrema=extrema,
        ...     fractional=True,
        ... )
        >>> ph = yt.PhasePlot.from_profile(profile)
        >>> ph.save()
        """
        obj = cls.__new__(cls)
        data_source = profile.data_source
        return cls._initialize_instance(
            obj, data_source, profile, fontsize, figure_size, shading
        )

    def annotate_text(self, xpos=0.0, ypos=0.0, text=None, **text_kwargs):
        r"""
        Allow the user to insert text onto the plot
        The x-position and y-position must be given as well as the text string.
        Add *text* tp plot at location *xpos*, *ypos* in plot coordinates
        (see example below).

        Parameters
        ----------
        xpos : float
          Position on plot in x-coordinates.
        ypos : float
          Position on plot in y-coordinates.
        text : str
          The text to insert onto the plot.
        **text_kwargs : dict
          Extra keyword arguments will be passed to matplotlib text instance

        >>> plot.annotate_text(1e-15, 5e4, "Hello YT")

        """
        for f in self.data_source._determine_fields(list(self.plots.keys())):
            if self.plots[f].figure is not None and text is not None:
                self.plots[f].axes.text(
                    xpos,
                    ypos,
                    text,
                    fontproperties=self._font_properties,
                    **text_kwargs,
                )
            self._plot_text[f] = text
            self._text_xpos[f] = xpos
            self._text_ypos[f] = ypos
            self._text_kwargs[f] = text_kwargs
        return self

    @validate_plot
    def save(self, name: str | None = None, suffix: str | None = None, mpl_kwargs=None):
        r"""
        Saves a 2d profile plot.

        Parameters
        ----------
        name : str, optional
            The output file keyword.
        suffix : string, optional
           Specify the image type by its suffix. If not specified, the output
           type will be inferred from the filename. Defaults to '.png'.
        mpl_kwargs : dict, optional
           A dict of keyword arguments to be passed to matplotlib.

        >>> plot.save(mpl_kwargs={"bbox_inches": "tight"})

        """
        names = []
        if not self._plot_valid:
            self._setup_plots()
        if mpl_kwargs is None:
            mpl_kwargs = {}
        if name is None:
            name = str(self.profile.ds)
        name = os.path.expanduser(name)
        xfn = self.profile.x_field
        yfn = self.profile.y_field
        if isinstance(xfn, tuple):
            xfn = xfn[1]
        if isinstance(yfn, tuple):
            yfn = yfn[1]
        for f in self.profile.field_data:
            _f = f
            if isinstance(f, tuple):
                _f = _f[1]
            middle = f"2d-Profile_{xfn}_{yfn}_{_f}"
            splitname = os.path.split(name)
            if splitname[0] != "" and not os.path.isdir(splitname[0]):
                os.makedirs(splitname[0])
            if os.path.isdir(name) and name != str(self.profile.ds):
                name = name + (os.sep if name[-1] != os.sep else "")
                name += str(self.profile.ds)

            new_name = validate_image_name(name, suffix)
            if new_name == name:
                for v in self.plots.values():
                    out_name = v.save(name, mpl_kwargs)
                    names.append(out_name)
                return names

            name = new_name
            prefix, suffix = os.path.splitext(name)
            name = f"{prefix}_{middle}{suffix}"

            names.append(name)
            self.plots[f].save(name, mpl_kwargs)
        return names

    @invalidate_plot
    def set_title(self, field, title):
        """Set a title for the plot.

        Parameters
        ----------
        field : str
            The z field of the plot to add the title.
        title : str
            The title to add.

        Examples
        --------

        >>> plot.set_title(("gas", "mass"), "This is a phase plot")
        """
        self.plot_title[self.data_source._determine_fields(field)[0]] = title
        return self

    @invalidate_plot
    def annotate_title(self, title):
        """Set a title for the plot.

        Parameters
        ----------
        title : str
            The title to add.

        Examples
        --------

        >>> plot.annotate_title("This is a phase plot")

        """
        for f in self._profile.field_data:
            self.plot_title[self.data_source._determine_fields(f)[0]] = title
        return self

    @invalidate_plot
    def reset_plot(self):
        self.plots = {}
        return self

    @invalidate_plot
    def set_log(self, field, log):
        """set a field to log or linear.

        Parameters
        ----------
        field : string
            the field to set a transform
        log : boolean
            Log on/off.
        """
        p = self._profile
        if field == "all":
            self.x_log = log
            self.y_log = log
            for field in p.field_data:
                self.z_log[field] = log
            self._profile_valid = False
        else:
            (field,) = self.profile.data_source._determine_fields([field])
            if field == p.x_field:
                self.x_log = log
                self._profile_valid = False
            elif field == p.y_field:
                self.y_log = log
                self._profile_valid = False
            elif field in p.field_data:
                super().set_log(field, log)
            else:
                raise KeyError(f"Field {field} not in phase plot!")
        return self

    @invalidate_plot
    def set_unit(self, field, unit):
        """Sets a new unit for the requested field

        Parameters
        ----------
        field : string
           The name of the field that is to be changed.

        unit : string or Unit object
           The name of the new unit.
        """
        fd = self.data_source._determine_fields(field)[0]
        if fd == self.profile.x_field:
            self.profile.set_x_unit(unit)
        elif fd == self.profile.y_field:
            self.profile.set_y_unit(unit)
        elif fd in self.profile.field_data.keys():
            self.profile.set_field_unit(field, unit)
            self.plots[field].norm_handler.display_units = unit
        else:
            raise KeyError(f"Field {field} not in phase plot!")
        return self

    @invalidate_plot
    @invalidate_profile
    def set_xlim(self, xmin=None, xmax=None):
        """Sets the limits of the x bin field

        Parameters
        ----------

        xmin : float or None
          The new x minimum in the current x-axis units.  Defaults to None,
          which leaves the xmin unchanged.

        xmax : float or None
          The new x maximum in the current x-axis units.  Defaults to None,
          which leaves the xmax unchanged.

        Examples
        --------

        >>> import yt
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> pp = yt.PhasePlot(ds.all_data(), "density", "temperature", ("gas", "mass"))
        >>> pp.set_xlim(1e-29, 1e-24)
        >>> pp.save()

        """
        p = self._profile
        if xmin is None:
            xmin = p.x_bins.min()
        elif not hasattr(xmin, "units"):
            xmin = self.ds.quan(xmin, p.x_bins.units)
        if xmax is None:
            xmax = p.x_bins.max()
        elif not hasattr(xmax, "units"):
            xmax = self.ds.quan(xmax, p.x_bins.units)
        self._xlim = (xmin, xmax)
        return self

    @invalidate_plot
    @invalidate_profile
    def set_ylim(self, ymin=None, ymax=None):
        """Sets the plot limits for the y bin field.

        Parameters
        ----------

        ymin : float or None
          The new y minimum in the current y-axis units.  Defaults to None,
          which leaves the ymin unchanged.

        ymax : float or None
          The new y maximum in the current y-axis units.  Defaults to None,
          which leaves the ymax unchanged.

        Examples
        --------

        >>> import yt
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> pp = yt.PhasePlot(
        ...     ds.all_data(),
        ...     ("gas", "density"),
        ...     ("gas", "temperature"),
        ...     ("gas", "mass"),
        ... )
        >>> pp.set_ylim(1e4, 1e6)
        >>> pp.save()

        """
        p = self._profile
        if ymin is None:
            ymin = p.y_bins.min()
        elif not hasattr(ymin, "units"):
            ymin = self.ds.quan(ymin, p.y_bins.units)
        if ymax is None:
            ymax = p.y_bins.max()
        elif not hasattr(ymax, "units"):
            ymax = self.ds.quan(ymax, p.y_bins.units)
        self._ylim = (ymin, ymax)
        return self

    def _recreate_profile(self):
        p = self._profile
        units = {p.x_field: str(p.x.units), p.y_field: str(p.y.units)}
        zunits = {field: str(p.field_units[field]) for field in p.field_units}
        extrema = {p.x_field: self._xlim, p.y_field: self._ylim}
        if self.x_log is not None or self.y_log is not None:
            logs = {}
        else:
            logs = None
        if self.x_log is not None:
            logs[p.x_field] = self.x_log
        if self.y_log is not None:
            logs[p.y_field] = self.y_log
        deposition = getattr(p, "deposition", None)
        additional_kwargs = {
            "accumulation": p.accumulation,
            "fractional": p.fractional,
            "deposition": deposition,
        }
        self._profile = create_profile(
            p.data_source,
            [p.x_field, p.y_field],
            list(p.field_map.values()),
            n_bins=[len(p.x_bins) - 1, len(p.y_bins) - 1],
            weight_field=p.weight_field,
            units=units,
            extrema=extrema,
            logs=logs,
            **additional_kwargs,
        )
        for field in zunits:
            self._profile.set_field_unit(field, zunits[field])
        self._profile_valid = True


class PhasePlotMPL(ImagePlotMPL):
    """A container for a single matplotlib figure and axes for a PhasePlot"""

    def __init__(
        self,
        x_data,
        y_data,
        data,
        x_scale,
        y_scale,
        figure_size,
        fontsize,
        figure,
        axes,
        cax,
        shading="nearest",
        *,
        norm_handler: NormHandler,
        colorbar_handler: ColorbarHandler,
    ):
        self._initfinished = False
        self._shading = shading
        self._setup_layout_constraints(figure_size, fontsize)

        # this line is added purely to prevent exact image comparison tests
        # to fail, but eventually we should embrace the change and
        # use similar values for PhasePlotMPL and WindowPlotMPL
        self._ax_text_size[0] *= 1.1 / 1.2  # TODO: remove this

        super().__init__(
            figure=figure,
            axes=axes,
            cax=cax,
            norm_handler=norm_handler,
            colorbar_handler=colorbar_handler,
        )

        self._init_image(x_data, y_data, data, x_scale, y_scale)

        self._initfinished = True

    def _init_image(
        self,
        x_data,
        y_data,
        image_data,
        x_scale,
        y_scale,
    ):
        """Store output of imshow in image variable"""
        norm = self.norm_handler.get_norm(image_data)
        self.image = None
        self.cb = None

        self.image = self.axes.pcolormesh(
            np.array(x_data),
            np.array(y_data),
            np.array(image_data.T),
            norm=norm,
            cmap=self.colorbar_handler.cmap,
            shading=self._shading,
        )

        self._set_axes()
        self.axes.set_xscale(x_scale)
        self.axes.set_yscale(y_scale)



# avoid deprecation warnings in numpy >= 2.0

import numpy as np

if hasattr(np, "trapezoid"):
    # np.trapz is deprecated in numpy 2.0 in favor of np.trapezoid
    trapezoid = np.trapezoid
else:
    trapezoid = np.trapz  # type: ignore[assignment] # noqa: NPY201



import numpy as np

import yt

# Open a dataset from when there's a lot of sloshing going on.

ds = yt.load("GasSloshingLowRes/sloshing_low_res_hdf5_plt_cnt_0350")

# Define the components of the gravitational acceleration vector field by
# taking the gradient of the gravitational potential
grad_fields = ds.add_gradient_fields(("gas", "gravitational_potential"))

# We don't need to do the same for the pressure field because yt already
# has pressure gradient fields. Now, define the "degree of hydrostatic
# equilibrium" field.


def _hse(field, data):
    # Remember that g is the negative of the potential gradient
    gx = -data["gas", "density"] * data["gas", "gravitational_potential_gradient_x"]
    gy = -data["gas", "density"] * data["gas", "gravitational_potential_gradient_y"]
    gz = -data["gas", "density"] * data["gas", "gravitational_potential_gradient_z"]
    hx = data["gas", "pressure_gradient_x"] - gx
    hy = data["gas", "pressure_gradient_y"] - gy
    hz = data["gas", "pressure_gradient_z"] - gz
    h = np.sqrt((hx * hx + hy * hy + hz * hz) / (gx * gx + gy * gy + gz * gz))
    return h


ds.add_field(
    ("gas", "HSE"),
    function=_hse,
    units="",
    take_log=False,
    display_name="Hydrostatic Equilibrium",
    sampling_type="cell",
)

# The gradient operator requires periodic boundaries.  This dataset has
# open boundary conditions.
ds.force_periodicity()

# Take a slice through the center of the domain
slc = yt.SlicePlot(ds, 2, [("gas", "density"), ("gas", "HSE")], width=(1, "Mpc"))

slc.save("hse")



import logging
import os
import re
import sys
import textwrap

import numpy as np
import pytest

from yt.config import ytcfg
from yt.loaders import load_sample
from yt.sample_data.api import (
    get_data_registry_table,
    get_download_cache_dir,
)
from yt.testing import requires_module_pytest
from yt.utilities.logger import ytLogger


@pytest.fixture()
def tmp_data_dir(tmp_path):
    pre_test_data_dir = ytcfg["yt", "test_data_dir"]
    ytcfg.set("yt", "test_data_dir", str(tmp_path))

    yield tmp_path

    ytcfg.set("yt", "test_data_dir", pre_test_data_dir)


@pytest.fixture()
def capturable_logger(caplog):
    """
    This set the minimal conditions to make pytest's caplog fixture usable.
    """

    propagate = ytLogger.propagate
    ytLogger.propagate = True

    with caplog.at_level(logging.INFO, "yt"):
        yield

    ytLogger.propagate = propagate


@requires_module_pytest("pandas", "pooch", "f90nml")
@pytest.mark.usefixtures("capturable_logger")
@pytest.mark.parametrize(
    "fn, archive, exact_loc, class_",
    [
        (
            "ToroShockTube",
            "ToroShockTube.tar.gz",
            "ToroShockTube/DD0001/data0001",
            "EnzoDataset",
        ),
        (
            "KeplerianDisk",
            "KeplerianDisk.tar.gz",
            "KeplerianDisk/disk.out1.00000.athdf",
            "AthenaPPDataset",
        ),
        # trying with an exact name as well
        (
            "KeplerianDisk/disk.out1.00000.athdf",
            "KeplerianDisk.tar.gz",
            "KeplerianDisk/disk.out1.00000.athdf",
            "AthenaPPDataset",
        ),
        # check this special case because it relies on implementations
        # details in the AMRVAC frontend (using parfiles)
        # and could easily fail to load. See GH PR #3343
        (
            "rmi_dust_2d",
            "rmi_dust_2d.tar.gz",
            "rmi_dust_2d/output0001.dat",
            "AMRVACDataset",
        ),
    ],
)
def test_load_sample_small_dataset(
    fn, archive, exact_loc, class_: str, tmp_data_dir, caplog
):
    cache_path = get_download_cache_dir()
    assert not cache_path.exists()

    ds = load_sample(fn, progressbar=False, timeout=30)
    assert type(ds).__name__ == class_
    assert not cache_path.exists()

    text = textwrap.dedent(
        f"""
            '{fn.replace('/', os.path.sep)}' is not available locally. Looking up online.
            Downloading from https://yt-project.org/data/{archive}
            Untaring downloaded file to '{str(tmp_data_dir)}'
        """
    ).strip("\n")
    expected = [("yt", 20, message) for message in text.split("\n")]
    assert caplog.record_tuples[:3] == expected

    caplog.clear()
    # loading a second time should not result in a download request
    ds2 = load_sample(fn)
    assert type(ds2).__name__ == class_

    assert caplog.record_tuples[0] == (
        "yt",
        20,
        f"Sample dataset found in '{os.path.join(str(tmp_data_dir), *exact_loc.split('/'))}'",
    )


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    # flakyness is probably due to Windows' infamous lack of time resolution
    # overall, this test doesn't seem worth it.
    reason="This test is flaky on Windows",
)
@requires_module_pytest("pandas", "pooch")
@pytest.mark.usefixtures("capturable_logger")
def test_load_sample_timeout(tmp_data_dir, caplog):
    from requests.exceptions import ConnectTimeout, ReadTimeout

    # note that requests is a direct dependency to pooch,
    # so we don't need to mark it in a decorator.
    with pytest.raises((ConnectTimeout, ReadTimeout)):
        load_sample("IsolatedGalaxy", progressbar=False, timeout=0.00001)


@requires_module_pytest("pandas", "requests")
@pytest.mark.xfail(reason="Registry is currently incomplete.")
def test_registry_integrity():
    reg = get_data_registry_table()
    assert not any(reg.isna())


@pytest.fixture()
def sound_subreg():
    # this selection is needed because the full dataset is incomplete
    # so we test only the values that can be parsed
    reg = get_data_registry_table()
    return reg[["size", "byte_size"]][reg["size"].notna()]


@requires_module_pytest("pandas", "requests")
def test_registry_byte_size_dtype(sound_subreg):
    from pandas import Int64Dtype

    assert sound_subreg["byte_size"].dtype == Int64Dtype()


@requires_module_pytest("pandas", "requests")
def test_registry_byte_size_sign(sound_subreg):
    np.testing.assert_array_less(0, sound_subreg["byte_size"])


@requires_module_pytest("pandas", "requests")
def test_unknown_filename():
    fake_name = "these_are_not_the_files_your_looking_for"
    with pytest.raises(ValueError, match=f"'{fake_name}' is not an available dataset."):
        load_sample(fake_name)


@requires_module_pytest("pandas", "requests")
def test_typo_filename():
    wrong_name = "Isolatedgalaxy"
    with pytest.raises(
        ValueError,
        match=re.escape(
            f"'{wrong_name}' is not an available dataset. Did you mean 'IsolatedGalaxy' ?"
        ),
    ):
        load_sample(wrong_name, timeout=1)


@pytest.fixture()
def fake_data_dir_in_a_vaccum(tmp_path):
    pre_test_data_dir = ytcfg["yt", "test_data_dir"]
    ytcfg.set("yt", "test_data_dir", "/does/not/exist")
    origin = os.getcwd()
    os.chdir(tmp_path)

    yield

    ytcfg.set("yt", "test_data_dir", pre_test_data_dir)
    os.chdir(origin)


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="can't figure out how to match the warning message in a cross-platform way",
)
@requires_module_pytest("pandas", "pooch")
@pytest.mark.usefixtures("fake_data_dir_in_a_vaccum")
def test_data_dir_broken():
    # check that load_sample still works if the test_data_dir
    # isn't properly set, in which case we should dl to the
    # cwd and issue a warning.
    msg = (
        r"Storage directory from yt config doesn't exist "
        r"\(currently set to '/does/not/exist'\)\. "
        r"Current working directory will be used instead\."
    )
    with pytest.warns(UserWarning, match=msg):
        load_sample("ToroShockTube")


def test_filename_none(capsys):
    assert load_sample() is None
    captured = capsys.readouterr()
    assert "yt.sample_data.api.get_data_registry_table" in captured.err



from numpy.testing import assert_allclose, assert_array_less, assert_raises

import yt
from yt.loaders import load
from yt.testing import fake_random_ds, requires_file, requires_module
from yt.utilities.exceptions import YTBoundsDefinitionError


def test_cic_deposit():
    ds = fake_random_ds(64, nprocs=8, particles=64**3)
    my_reg = ds.arbitrary_grid(
        ds.domain_left_edge, ds.domain_right_edge, dims=[1, 800, 800]
    )
    f = ("deposit", "all_cic")
    assert_raises(YTBoundsDefinitionError, my_reg.__getitem__, f)


RAMSES = "output_00080/info_00080.txt"
RAMSES_small = "ramses_new_format/output_00002/info_00002.txt"
ISOGAL = "IsolatedGalaxy/galaxy0030/galaxy0030"


@requires_file(RAMSES)
def test_one_zone_octree_deposit():
    ds = load(RAMSES)

    # Get a sphere centred on the main halo
    hpos = ds.arr(
        [0.5215110772898429, 0.5215110772898429, 0.5215110772898429], "code_length"
    )
    hrvir = ds.quan(0.042307235300540924, "Mpc")

    sp = ds.sphere(hpos, hrvir * 10)
    assert sp["deposit", "io_cic"].shape == (1,)


@requires_module("h5py")
@requires_file(RAMSES)
@requires_file(ISOGAL)
def test_mesh_sampling():
    for fn in (RAMSES, ISOGAL):
        ds = yt.load(fn)
        ds.add_mesh_sampling_particle_field(("index", "x"), ptype="all")
        ds.add_mesh_sampling_particle_field(("index", "dx"), ptype="all")

        dx = ds.r["all", "cell_index_dx"]
        xc = ds.r["all", "cell_index_x"]
        xp = ds.r["all", "particle_position_x"]

        dist = xp - xc

        assert_array_less(dist, dx)
        assert_array_less(-dist, dx)


@requires_module("h5py")
@requires_file(RAMSES)
@requires_file(ISOGAL)
def test_mesh_sampling_for_filtered_particles():
    for fn in (RAMSES, ISOGAL):
        ds = yt.load(fn)

        @yt.particle_filter(requires=["particle_position_x"], filtered_type="io")
        def left(pfilter, data):
            return (
                data[pfilter.filtered_type, "particle_position_x"].to("code_length")
                < 0.5
            )

        ds.add_particle_filter("left")

        for f in (("index", "x"), ("index", "dx"), ("gas", "density")):
            ds.add_mesh_sampling_particle_field(f, ptype="io")
            ds.add_mesh_sampling_particle_field(f, ptype="left")

        data_sources = (ds.all_data(), ds.box([0] * 3, [0.1] * 3))

        def test_source(ptype, src):
            # Test accessing
            src[ptype, "cell_index_x"]
            src[ptype, "cell_index_dx"]
            src[ptype, "cell_gas_density"]

        for ptype in ("io", "left"):
            for src in data_sources:
                test_source(ptype, src)


@requires_file(RAMSES)
def test_mesh_sampling_with_indexing():
    # Access with index caching
    ds = yt.load(RAMSES)
    ds.add_mesh_sampling_particle_field(("gas", "density"), ptype="all")

    ad = ds.all_data()
    ad["all", "cell_index"]
    v1 = ad["all", "cell_gas_density"]

    # Access with no index caching
    ds = yt.load(RAMSES)
    ds.add_mesh_sampling_particle_field(("gas", "density"), ptype="all")

    ad = ds.all_data()
    v2 = ad["all", "cell_gas_density"]

    # Check same answer is returned
    assert_allclose(v1, v2)


@requires_file(RAMSES_small)
def test_mesh_sampling_vs_field_value_at_point():
    all_ds = (fake_random_ds(ndims=3, particles=500), yt.load(RAMSES_small))

    for ds in all_ds:
        ds.add_mesh_sampling_particle_field(("gas", "density"), ptype="all")

        val = ds.r["all", "cell_gas_density"]
        ref = ds.find_field_values_at_points(
            ("gas", "density"), ds.r["all", "particle_position"]
        )

        assert_allclose(val, ref)



"""
Fixed interpolator includes



"""


cimport numpy as np


cdef extern from "fixed_interpolator.hpp":
    np.float64_t fast_interpolate(int ds[3], int ci[3], np.float64_t dp[3],
                                  np.float64_t *data) noexcept nogil
    np.float64_t offset_interpolate(int ds[3], np.float64_t dp[3],
                                    np.float64_t *data) noexcept nogil
    np.float64_t trilinear_interpolate(int ds[3], int ci[3], np.float64_t dp[3],
                                       np.float64_t *data) noexcept nogil
    void eval_gradient(int ds[3], np.float64_t dp[3], np.float64_t *data,
                       np.float64_t grad[3]) noexcept nogil
    void offset_fill(int *ds, np.float64_t *data, np.float64_t *gridval) noexcept nogil
    void vertex_interp(np.float64_t v1, np.float64_t v2, np.float64_t isovalue,
                       np.float64_t vl[3], np.float64_t dds[3],
                       np.float64_t x, np.float64_t y, np.float64_t z,
                       int vind1, int vind2) noexcept nogil



import os
import struct

import numpy as np

from yt.data_objects.static_output import ParticleFile
from yt.fields.field_info_container import FieldInfoContainer
from yt.frontends.sph.data_structures import SPHDataset, SPHParticleIndex
from yt.funcs import only_on_root
from yt.geometry.geometry_handler import Index
from yt.utilities.chemical_formulas import compute_mu
from yt.utilities.cosmology import Cosmology
from yt.utilities.fortran_utils import read_record
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py

from .definitions import (
    SNAP_FORMAT_2_OFFSET,
    gadget_field_specs,
    gadget_header_specs,
    gadget_ptype_specs,
)
from .fields import GadgetFieldInfo


def _fix_unit_ordering(unit):
    if isinstance(unit[0], str):
        unit = unit[1], unit[0]
    return unit


def _byte_swap_32(x):
    return struct.unpack(">I", struct.pack("<I", x))[0]


class GadgetBinaryHeader:
    """A convenient interface to Gadget binary header.

    This is a helper class to facilitate the main dataset and IO classes.
    The main usage is through the GadgetDataset._header attribute. It is also
    used stand-alone in GadgetDataset._is_valid method.
    """

    _placeholder_keys = ["unused", "empty"]

    def __init__(self, filename, header_spec):
        self.filename = filename
        if isinstance(header_spec, str):
            header_spec = [header_spec]
        self.spec = [
            GadgetDataset._setup_binary_spec(hs, gadget_header_specs)
            for hs in header_spec
        ]

    @property
    def float_type(self):
        """Determine whether the floats are single or double precision."""
        gformat, endianswap = self.gadget_format
        # We do not currently support double precision snapshot format 2
        if gformat == 2:
            return "f4"
        else:
            hvals = self.value
            np0 = sum(hvals["Npart"])
            with self.open() as f:
                f.seek(self.position_offset)
                # Calculate particle number assuming single precision
                np1 = struct.unpack(endianswap + "I", f.read(4))[0] / (4 * 3)
            if np1 == np0:
                return "f4"
            elif np1 == 2 * np0:
                return "f8"
            else:
                raise RuntimeError(
                    "Gadget snapshot file is likely corrupted! "
                    "Cannot determine float type."
                )

    @property
    def gadget_format(self):
        """Determine Gadget snapshot format and endianness.

        The difference between Gadget snapshot format 1 and 2 can be told from
        the first 4 bytes of the file. For format 1, it's the header size. For
        format 2, it's always 8.
        """
        first_header_size = self.size[0]
        # Read the first 4 bytes assuming little endian int32
        with self.open() as f:
            (rhead,) = struct.unpack("<I", f.read(4))
        # Format 1?
        if rhead == first_header_size:
            return 1, "<"
        elif rhead == _byte_swap_32(first_header_size):
            return 1, ">"
        # Format 2?
        elif rhead == 8:
            return 2, "<"
        elif rhead == _byte_swap_32(8):
            return 2, ">"
        else:
            raise RuntimeError(
                "Gadget snapshot file is likely corrupted! "
                f"The first 4 bytes represent {rhead} (as little endian int32). "
                f"But we are looking for {first_header_size} (for format 1) "
                "or 8 (for format 2)."
            )

    @property
    def position_offset(self):
        """Offset to the position block."""
        n_header = len(self.size)
        offset = 8 * n_header + sum(self.size)
        if self.gadget_format[0] == 2:
            offset += SNAP_FORMAT_2_OFFSET * (n_header + 1)
        return offset

    @property
    def size(self):
        """Header size in bytes."""

        def single_header_size(single_header_spec):
            return sum(
                field[1] * np.dtype(field[2]).itemsize for field in single_header_spec
            )

        return [single_header_size(spec) for spec in self.spec]

    @property
    def value(self):
        """Header values as a dictionary."""
        # The entries in this header are capitalized and named to match Table 4
        # in the GADGET-2 user guide.
        gformat, endianswap = self.gadget_format
        # Read header
        with self.open() as f:
            hvals = {}
            for spec in self.spec:
                if gformat == 2:
                    f.seek(f.tell() + SNAP_FORMAT_2_OFFSET)
                hvals_new = read_record(f, spec, endian=endianswap)
                hvals.update(hvals_new)
        # Remove placeholder keys
        for key in self._placeholder_keys:
            if key in hvals:
                del hvals[key]
        # Convert length 1 list to scalar value
        for i in hvals:
            if len(hvals[i]) == 1:
                hvals[i] = hvals[i][0]
        return hvals

    def open(self):
        """Open snapshot file."""
        for filename in [self.filename, self.filename + ".0"]:
            if os.path.exists(filename):
                return open(filename, "rb")
        raise FileNotFoundError(f"Snapshot file {self.filename} does not exist.")

    def validate(self):
        """Validate data integrity."""
        try:
            self.open().close()
            self.gadget_format
            self.float_type
        except Exception:
            return False
        return True


class GadgetBinaryFile(ParticleFile):
    def __init__(self, ds, io, filename, file_id, range=None):
        header = GadgetBinaryHeader(filename, ds._header.spec)
        self.header = header.value
        self._position_offset = header.position_offset
        with header.open() as f:
            self._file_size = f.seek(0, os.SEEK_END)

        super().__init__(ds, io, filename, file_id, range)

    def _calculate_offsets(self, field_list, pcounts):
        # Note that we ignore pcounts here because it's the global count.  We
        # just want the local count, which we store here.
        self.field_offsets = self.io._calculate_field_offsets(
            field_list,
            self.header["Npart"].copy(),
            self._position_offset,
            self.start,
            self._file_size,
        )


class GadgetBinaryIndex(SPHParticleIndex):
    def __init__(self, ds, dataset_type):
        super().__init__(ds, dataset_type)
        self._initialize_index()

    def _initialize_index(self):
        # Normally this function is called during field detection. We call it
        # here because we need to know which fields exist on-disk so that we can
        # read in the smoothing lengths for SPH data before we construct the
        # Morton bitmaps.
        self._detect_output_fields()
        super()._initialize_index()

    def _initialize_frontend_specific(self):
        super()._initialize_frontend_specific()
        self.io._float_type = self.ds._header.float_type


class GadgetDataset(SPHDataset):
    _index_class: type[Index] = GadgetBinaryIndex
    _file_class: type[ParticleFile] = GadgetBinaryFile
    _field_info_class: type[FieldInfoContainer] = GadgetFieldInfo
    _particle_mass_name = "Mass"
    _particle_coordinates_name = "Coordinates"
    _particle_velocity_name = "Velocities"
    _sph_ptypes = ("Gas",)
    _suffix = ""

    def __init__(
        self,
        filename,
        dataset_type="gadget_binary",
        additional_fields=(),
        unit_base=None,
        index_order=None,
        index_filename=None,
        kdtree_filename=None,
        kernel_name=None,
        bounding_box=None,
        header_spec="default",
        field_spec="default",
        ptype_spec="default",
        long_ids=False,
        units_override=None,
        mean_molecular_weight=None,
        header_offset=0,
        unit_system="cgs",
        use_dark_factor=False,
        w_0=-1.0,
        w_a=0.0,
        default_species_fields=None,
    ):
        if self._instantiated:
            return
        # Check if filename is a directory
        if os.path.isdir(filename):
            # Get the .0 snapshot file. We know there's only 1 and it's valid since we
            # came through _is_valid in load()
            for f in os.listdir(filename):
                fname = os.path.join(filename, f)
                fext = os.path.splitext(fname)[-1]
                if (
                    (".0" in f)
                    and (fext not in {".ewah", ".kdtree"})
                    and os.path.isfile(fname)
                ):
                    filename = os.path.join(filename, f)
                    break
        self._header = GadgetBinaryHeader(filename, header_spec)
        header_size = self._header.size
        if header_size != [256]:
            only_on_root(
                mylog.warning,
                "Non-standard header size is detected! "
                "Gadget-2 standard header is 256 bytes, but yours is %s. "
                "Make sure a non-standard header is actually expected. "
                "Otherwise something is wrong, "
                "and you might want to check how the dataset is loaded. "
                "Further information about header specification can be found in "
                "https://yt-project.org/docs/dev/examining/loading_data.html#header-specification.",
                header_size,
            )
        self._field_spec = self._setup_binary_spec(field_spec, gadget_field_specs)
        self._ptype_spec = self._setup_binary_spec(ptype_spec, gadget_ptype_specs)
        self.storage_filename = None
        if long_ids:
            self._id_dtype = "u8"
        else:
            self._id_dtype = "u4"
        self.long_ids = long_ids
        self.header_offset = header_offset
        if unit_base is not None and "UnitLength_in_cm" in unit_base:
            # We assume this is comoving, because in the absence of comoving
            # integration the redshift will be zero.
            unit_base["cmcm"] = 1.0 / unit_base["UnitLength_in_cm"]
        self._unit_base = unit_base
        if bounding_box is not None:
            # This ensures that we know a bounding box has been applied
            self._domain_override = True
            bbox = np.array(bounding_box, dtype="float64")
            if bbox.shape == (2, 3):
                bbox = bbox.transpose()
            self.domain_left_edge = bbox[:, 0]
            self.domain_right_edge = bbox[:, 1]
        else:
            self.domain_left_edge = self.domain_right_edge = None
        if units_override is not None:
            raise RuntimeError(
                "units_override is not supported for GadgetDataset. "
                + "Use unit_base instead."
            )

        # Set dark energy parameters before cosmology object is created
        self.use_dark_factor = use_dark_factor
        self.w_0 = w_0
        self.w_a = w_a

        super().__init__(
            filename,
            dataset_type=dataset_type,
            unit_system=unit_system,
            index_order=index_order,
            index_filename=index_filename,
            kdtree_filename=kdtree_filename,
            kernel_name=kernel_name,
            default_species_fields=default_species_fields,
        )
        if self.cosmological_simulation:
            self.time_unit.convert_to_units("s/h")
            self.length_unit.convert_to_units("kpccm/h")
            self.mass_unit.convert_to_units("g/h")
        else:
            self.time_unit.convert_to_units("s")
            self.length_unit.convert_to_units("kpc")
            self.mass_unit.convert_to_units("Msun")
        if mean_molecular_weight is None:
            self.mu = compute_mu(self.default_species_fields)
        else:
            self.mu = mean_molecular_weight

    @classmethod
    def _setup_binary_spec(cls, spec, spec_dict):
        if isinstance(spec, str):
            _hs = ()
            for hs in spec.split("+"):
                _hs += spec_dict[hs]
            spec = _hs
        return spec

    def __str__(self):
        return os.path.basename(self.parameter_filename).split(".")[0]

    def _get_hvals(self):
        self.gen_hsmls = False
        return self._header.value

    def _parse_parameter_file(self):
        hvals = self._get_hvals()

        self.dimensionality = 3
        self.refine_by = 2
        self.parameters["HydroMethod"] = "sph"
        # Set standard values

        # We may have an overridden bounding box.
        if self.domain_left_edge is None and hvals["BoxSize"] != 0:
            self.domain_left_edge = np.zeros(3, "float64")
            self.domain_right_edge = np.ones(3, "float64") * hvals["BoxSize"]

        self.domain_dimensions = np.ones(3, "int32")
        self._periodicity = (True, True, True)

        self.current_redshift = hvals.get("Redshift", 0.0)
        if "Redshift" not in hvals:
            mylog.info("Redshift is not set in Header. Assuming z=0.")

        if "OmegaLambda" in hvals:
            self.omega_lambda = hvals["OmegaLambda"]
            self.omega_matter = hvals["Omega0"]
            self.hubble_constant = hvals["HubbleParam"]
            self.cosmological_simulation = self.omega_lambda != 0.0
        else:
            # If these are not set it is definitely not a cosmological dataset.
            self.omega_lambda = 0.0
            self.omega_matter = 0.0  # Just in case somebody asks for it.
            # omega_matter has been changed to 0.0 for consistency with
            # non-cosmological frontends
            self.cosmological_simulation = 0
            # Hubble is set below for Omega Lambda = 0.

        # According to the Gadget manual, OmegaLambda will be zero for
        # non-cosmological datasets.  However, it may be the case that
        # individuals are running cosmological simulations *without* Lambda, in
        # which case we may be doing something incorrect here.
        # It may be possible to deduce whether ComovingIntegration is on
        # somehow, but opinions on this vary.
        if not self.cosmological_simulation:
            mylog.info("Omega Lambda is 0.0, so we are turning off Cosmology.")
            self.hubble_constant = 1.0  # So that scaling comes out correct
            self.current_redshift = 0.0
            # This may not be correct.
            self.current_time = hvals["Time"]
        else:
            # Now we calculate our time based on the cosmology, because in
            # ComovingIntegration hvals["Time"] will in fact be the expansion
            # factor, not the actual integration time, so we re-calculate
            # global time from our Cosmology.
            cosmo = Cosmology(
                hubble_constant=self.hubble_constant,
                omega_matter=self.omega_matter,
                omega_lambda=self.omega_lambda,
            )
            self.current_time = cosmo.lookback_time(self.current_redshift, 1e6)
            mylog.info(
                "Calculating time from %0.3e to be %0.3e seconds",
                hvals["Time"],
                self.current_time,
            )
        self.parameters = hvals

        basename, _, _ = self.basename.partition(".")
        prefix = os.path.join(self.directory, basename)

        if hvals["NumFiles"] > 1:
            for t in (
                f"{prefix}.%(num)s{self._suffix}",
                f"{prefix}.gad.%(num)s{self._suffix}",
            ):
                if os.path.isfile(t % {"num": 0}):
                    self.filename_template = t
                    break
            else:
                raise RuntimeError("Could not determine correct data file template.")
        else:
            self.filename_template = self.parameter_filename

        self.file_count = hvals["NumFiles"]

    def _set_code_unit_attributes(self):
        # If no units passed in by user, set a sane default (Gadget-2 users
        # guide).
        if self._unit_base is None:
            if self.cosmological_simulation == 1:
                only_on_root(
                    mylog.info, "Assuming length units are in kpc/h (comoving)"
                )
                self._unit_base = {"length": (1.0, "kpccm/h")}
            else:
                only_on_root(mylog.info, "Assuming length units are in kpc (physical)")
                self._unit_base = {"length": (1.0, "kpc")}

        # If units passed in by user, decide what to do about
        # co-moving and factors of h
        unit_base = self._unit_base or {}
        if "length" in unit_base:
            length_unit = unit_base["length"]
        elif "UnitLength_in_cm" in unit_base:
            if self.cosmological_simulation == 0:
                length_unit = (unit_base["UnitLength_in_cm"], "cm")
            else:
                length_unit = (unit_base["UnitLength_in_cm"], "cmcm/h")
        else:
            raise RuntimeError
        length_unit = _fix_unit_ordering(length_unit)
        self.length_unit = self.quan(length_unit[0], length_unit[1])

        unit_base = self._unit_base or {}

        if self.cosmological_simulation:
            # see http://www.mpa-garching.mpg.de/gadget/gadget-list/0113.html
            # for why we need to include a factor of square root of the
            # scale factor
            vel_units = "cm/s * sqrt(a)"
        else:
            vel_units = "cm/s"

        if "velocity" in unit_base:
            velocity_unit = unit_base["velocity"]
        elif "UnitVelocity_in_cm_per_s" in unit_base:
            velocity_unit = (unit_base["UnitVelocity_in_cm_per_s"], vel_units)
        else:
            velocity_unit = (1e5, vel_units)
        velocity_unit = _fix_unit_ordering(velocity_unit)
        self.velocity_unit = self.quan(velocity_unit[0], velocity_unit[1])

        # We set hubble_constant = 1.0 for non-cosmology, so this is safe.
        # Default to 1e10 Msun/h if mass is not specified.
        if "mass" in unit_base:
            mass_unit = unit_base["mass"]
        elif "UnitMass_in_g" in unit_base:
            if self.cosmological_simulation == 0:
                mass_unit = (unit_base["UnitMass_in_g"], "g")
            else:
                mass_unit = (unit_base["UnitMass_in_g"], "g/h")
        else:
            # Sane default
            mass_unit = (1e10, "Msun/h")
        mass_unit = _fix_unit_ordering(mass_unit)
        self.mass_unit = self.quan(mass_unit[0], mass_unit[1])
        if self.cosmological_simulation:
            # self.velocity_unit is the unit to rescale on-disk velocities, The
            # actual internal velocity unit is really in comoving units
            # since the time unit is derived from the internal velocity unit, we
            # infer the internal velocity unit here and name it vel_unit
            #
            # see http://www.mpa-garching.mpg.de/gadget/gadget-list/0113.html
            if "velocity" in unit_base:
                vel_unit = unit_base["velocity"]
            elif "UnitVelocity_in_cm_per_s" in unit_base:
                vel_unit = (unit_base["UnitVelocity_in_cm_per_s"], "cmcm/s")
            else:
                vel_unit = (1, "kmcm/s")
            vel_unit = self.quan(*vel_unit)
        else:
            vel_unit = self.velocity_unit
        self.time_unit = self.length_unit / vel_unit

        if "specific_energy" in unit_base:
            specific_energy_unit = unit_base["specific_energy"]
        elif "UnitEnergy_in_cgs" in unit_base and "UnitMass_in_g" in unit_base:
            specific_energy_unit = (
                unit_base["UnitEnergy_in_cgs"] / unit_base["UnitMass_in_g"]
            )
            specific_energy_unit = (specific_energy_unit, "(cm/s)**2")
        else:
            # Sane default
            specific_energy_unit = (1, "(km/s) ** 2")
        specific_energy_unit = _fix_unit_ordering(specific_energy_unit)
        self.specific_energy_unit = self.quan(*specific_energy_unit)

        if "magnetic" in unit_base:
            magnetic_unit = unit_base["magnetic"]
        elif "UnitMagneticField_in_gauss" in unit_base:
            magnetic_unit = (unit_base["UnitMagneticField_in_gauss"], "gauss")
        else:
            # Sane default
            magnetic_unit = (1.0, "gauss")
        magnetic_unit = _fix_unit_ordering(magnetic_unit)
        self.magnetic_unit = self.quan(*magnetic_unit)

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        header_spec = kwargs.get("header_spec", "default")
        # Check to see if passed filename is a directory. If so, use it to get
        # the .0 snapshot file. Make sure there's only one such file, otherwise
        # there's an ambiguity about which file the user wants. Ignore ewah files
        if os.path.isdir(filename):
            valid_files = []
            for f in os.listdir(filename):
                fname = os.path.join(filename, f)
                if (".0" in f) and (".ewah" not in f) and os.path.isfile(fname):
                    valid_files.append(f)
            if len(valid_files) == 0:
                return False
            elif len(valid_files) > 1:
                return False
            else:
                validated_file = os.path.join(filename, valid_files[0])
        else:
            validated_file = filename
        header = GadgetBinaryHeader(validated_file, header_spec)
        return header.validate()


class GadgetHDF5File(ParticleFile):
    pass


class GadgetHDF5Dataset(GadgetDataset):
    _load_requirements = ["h5py"]
    _file_class = GadgetHDF5File
    _index_class = SPHParticleIndex
    _field_info_class: type[FieldInfoContainer] = GadgetFieldInfo
    _particle_mass_name = "Masses"
    _sph_ptypes = ("PartType0",)
    _suffix = ".hdf5"

    def __init__(
        self,
        filename,
        dataset_type="gadget_hdf5",
        unit_base=None,
        index_order=None,
        index_filename=None,
        kernel_name=None,
        bounding_box=None,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
    ):
        self.storage_filename = None
        filename = os.path.abspath(filename)
        if units_override is not None:
            raise RuntimeError(
                "units_override is not supported for GadgetHDF5Dataset. "
                "Use unit_base instead."
            )
        super().__init__(
            filename,
            dataset_type,
            unit_base=unit_base,
            index_order=index_order,
            index_filename=index_filename,
            kernel_name=kernel_name,
            bounding_box=bounding_box,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )

    def _get_hvals(self):
        handle = h5py.File(self.parameter_filename, mode="r")
        hvals = {}
        hvals.update((str(k), v) for k, v in handle["/Header"].attrs.items())
        # Compat reasons.
        hvals["NumFiles"] = hvals["NumFilesPerSnapshot"]
        hvals["Massarr"] = hvals["MassTable"]
        sph_ptypes = [ptype for ptype in self._sph_ptypes if ptype in handle]
        if sph_ptypes:
            self.gen_hsmls = "SmoothingLength" not in handle[sph_ptypes[0]]
        else:
            self.gen_hsmls = False
        # Later versions of Gadget and its derivatives have a "Parameters"
        # group in the HDF5 file.
        if "Parameters" in handle:
            hvals.update((str(k), v) for k, v in handle["/Parameters"].attrs.items())
        handle.close()

        # ensure that 1-element arrays are reduced to scalars
        updated_hvals = {}
        for hvalname, value in hvals.items():
            if isinstance(value, np.ndarray) and value.size == 1:
                mylog.info("Reducing single-element array %s to scalar.", hvalname)
                updated_hvals[hvalname] = value.item()
        hvals.update(updated_hvals)

        return hvals

    def _get_uvals(self):
        handle = h5py.File(self.parameter_filename, mode="r")
        uvals = {}
        uvals.update((str(k), v) for k, v in handle["/Units"].attrs.items())
        handle.close()
        return uvals

    def _set_owls_eagle(self):
        self.dimensionality = 3
        self.refine_by = 2
        self.parameters["HydroMethod"] = "sph"

        self._unit_base = self._get_uvals()
        self._unit_base["cmcm"] = 1.0 / self._unit_base["UnitLength_in_cm"]

        self.current_redshift = self.parameters["Redshift"]
        self.omega_lambda = self.parameters["OmegaLambda"]
        self.omega_matter = self.parameters["Omega0"]
        self.hubble_constant = self.parameters["HubbleParam"]

        if self.domain_left_edge is None and self.parameters["BoxSize"] != 0:
            self.domain_left_edge = np.zeros(3, "float64")
            self.domain_right_edge = np.ones(3, "float64") * self.parameters["BoxSize"]

        self.domain_dimensions = np.ones(3, "int32")

        self.cosmological_simulation = 1
        self._periodicity = (True, True, True)

        prefix = os.path.abspath(
            os.path.join(
                os.path.dirname(self.parameter_filename),
                os.path.basename(self.parameter_filename).split(".", 1)[0],
            )
        )

        suffix = self.parameter_filename.rsplit(".", 1)[-1]
        if self.parameters["NumFiles"] > 1:
            self.filename_template = f"{prefix}.%(num)i.{suffix}"
        else:
            self.filename_template = self.parameter_filename

        self.file_count = self.parameters["NumFilesPerSnapshot"]

    def _set_owls_eagle_units(self):
        # note the contents of the HDF5 Units group are in _unit_base
        # note the velocity stored on disk is sqrt(a) dx/dt
        # physical velocity [cm/s] = a dx/dt
        # = sqrt(a) * velocity_on_disk * UnitVelocity_in_cm_per_s
        self.length_unit = self.quan(self._unit_base["UnitLength_in_cm"], "cmcm/h")
        self.mass_unit = self.quan(self._unit_base["UnitMass_in_g"], "g/h")
        self.velocity_unit = self.quan(
            self._unit_base["UnitVelocity_in_cm_per_s"], "cm/s * sqrt(a)"
        )
        self.time_unit = self.quan(self._unit_base["UnitTime_in_s"], "s/h")

        specific_energy_unit_cgs = (
            self._unit_base["UnitEnergy_in_cgs"] / self._unit_base["UnitMass_in_g"]
        )
        self.specific_energy_unit = self.quan(specific_energy_unit_cgs, "(cm/s)**2")

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        need_groups = ["Header"]
        veto_groups = ["FOF", "Group", "Subhalo"]
        valid = True
        try:
            fh = h5py.File(filename, mode="r")
            valid = all(ng in fh["/"] for ng in need_groups) and not any(
                vg in fh["/"] for vg in veto_groups
            )
            fh.close()
        except Exception:
            valid = False
            pass

        try:
            fh = h5py.File(filename, mode="r")
            valid = fh["Header"].attrs["Code"].decode("utf-8") != "SWIFT"
            fh.close()
        except (OSError, KeyError):
            pass

        return valid



import numpy as np

import yt
from yt.visualization.volume_rendering.api import Scene, create_volume_source

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# create a scene and add volume sources to it

sc = Scene()

# Add density

field = "density"

vol = create_volume_source(ds, field=field)
vol.use_ghost_zones = True

tf = yt.ColorTransferFunction([-28, -25])
tf.clear()
tf.add_layers(4, 0.02, alpha=np.logspace(-3, -1, 4), colormap="winter")

vol.set_transfer_function(tf)
sc.add_source(vol)

# Add temperature

field = "temperature"

vol2 = create_volume_source(ds, field=field)
vol2.use_ghost_zones = True

tf = yt.ColorTransferFunction([4.5, 7.5])
tf.clear()
tf.add_layers(4, 0.02, alpha=np.logspace(-0.2, 0, 4), colormap="autumn")

vol2.set_transfer_function(tf)
sc.add_source(vol2)

# setup the camera

cam = sc.add_camera(ds, lens_type="perspective")
cam.resolution = (1600, 900)
cam.zoom(20.0)

# Render the image.

sc.render()

sc.save_annotated(
    "render_two_fields_tf.png",
    sigma_clip=6.0,
    tf_rect=[0.88, 0.15, 0.03, 0.8],
    render=False,
)



import numpy as np

from yt.data_objects.profiles import create_profile
from yt.testing import fake_particle_ds, fake_random_ds


def test_update_data_grid():
    ds = fake_random_ds(64, nprocs=8)
    ds.index
    dims = (32, 32, 32)
    grid_data = [
        {"temperature": np.random.uniform(size=dims)} for i in range(ds.index.num_grids)
    ]
    ds.index.update_data(grid_data)
    prj = ds.proj(("gas", "temperature"), 2)
    prj["gas", "temperature"]
    dd = ds.all_data()
    profile = create_profile(dd, ("gas", "density"), ("gas", "temperature"), 10)
    profile["gas", "temperature"]


def test_update_data_particle():
    npart = 100
    ds = fake_particle_ds(npart=npart)
    part_data = {"temperature": np.random.rand(npart)}
    ds.index.update_data(part_data)
    assert ("io", "temperature") in ds.field_list
    dd = ds.all_data()
    dd["io", "temperature"]



from numpy.testing import assert_equal

import yt
from yt.frontends.cholla.api import ChollaDataset
from yt.testing import requires_file, requires_module
from yt.utilities.answer_testing.framework import (
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

_fields = (
    ("gas", "temperature"),
    ("gas", "density"),
)

ChollaSimple = "ChollaSimple/0.h5"


@requires_module("h5py")
@requires_file(ChollaSimple)
def test_ChollaDataset():
    assert isinstance(data_dir_load(ChollaSimple), ChollaDataset)


@requires_module("h5py")
@requires_file(ChollaSimple)
def test_ChollaSimple_fields():
    expected_fields = [
        "Energy",
        "GasEnergy",
        "density",
        "momentum_x",
        "momentum_y",
        "momentum_z",
        "scalar0",
    ]

    ds = yt.load(ChollaSimple)
    assert_equal(str(ds), "0.h5")
    ad = ds.all_data()

    # Check all the expected fields exist and can be accessed
    for field in expected_fields:
        assert ("cholla", field) in ds.field_list

        # test that field access works
        ad["cholla", field]


@requires_module("h5py")
@requires_file(ChollaSimple)
def test_ChollaSimple_derived_fields():
    expected_derived_fields = [
        "density",
        "momentum_x",
        "momentum_y",
        "momentum_z",
        "metallicity",
    ]

    ds = yt.load(ChollaSimple)
    ad = ds.all_data()

    # Check all the expected fields exist and can be accessed
    for field in expected_derived_fields:
        assert ("gas", field) in ds.derived_field_list

        # test that field access works
        ad["gas", field]


_fields_chollasimple = (
    ("cholla", "GasEnergy"),
    ("gas", "temperature"),
    ("gas", "density"),
    ("gas", "metallicity"),
)


@requires_module("h5py")
@requires_ds(ChollaSimple)
def test_cholla_data():
    ds = data_dir_load(ChollaSimple)
    assert_equal(str(ds), "0.h5")
    for test in small_patch_amr(
        ds, _fields_chollasimple, input_center="c", input_weight="ones"
    ):
        test_cholla_data.__name__ = test.description
        yield test



import glob
import weakref
from collections import defaultdict
from functools import cached_property, partial

import numpy as np

from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
)
from yt.data_objects.static_output import (
    ParticleDataset,
    ParticleFile,
)
from yt.frontends.ytdata.data_structures import SavedDataset
from yt.funcs import parse_h5_attr
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.utilities.on_demand_imports import _h5py as h5py

from .fields import YTHaloCatalogFieldInfo, YTHaloCatalogHaloFieldInfo


class HaloCatalogFile(ParticleFile):
    """
    Base class for data files of halo catalog datasets.

    This is mainly here to correct for periodicity when
    reading particle positions.
    """

    def __init__(self, ds, io, filename, file_id, frange):
        super().__init__(ds, io, filename, file_id, frange)

    def _read_particle_positions(self, ptype, f=None):
        raise NotImplementedError

    def _get_particle_positions(self, ptype, f=None):
        pcount = self.total_particles[ptype]
        if pcount == 0:
            return None

        # Correct for periodicity.
        dle = self.ds.domain_left_edge.to("code_length").v
        dw = self.ds.domain_width.to("code_length").v
        pos = self._read_particle_positions(ptype, f=f)
        si, ei = self.start, self.end
        if None not in (si, ei):
            pos = pos[si:ei]

        np.subtract(pos, dle, out=pos)
        np.mod(pos, dw, out=pos)
        np.add(pos, dle, out=pos)

        return pos


class YTHaloCatalogFile(HaloCatalogFile):
    """
    Data file class for the YTHaloCatalogDataset.
    """

    def __init__(self, ds, io, filename, file_id, frange):
        with h5py.File(filename, mode="r") as f:
            self.header = {field: parse_h5_attr(f, field) for field in f.attrs.keys()}
            pids = f.get("particles/ids")
            self.total_ids = 0 if pids is None else pids.size
            self.group_length_sum = self.total_ids
        super().__init__(ds, io, filename, file_id, frange)

    def _read_particle_positions(self, ptype, f=None):
        """
        Read all particle positions in this file.
        """

        if f is None:
            close = True
            f = h5py.File(self.filename, mode="r")
        else:
            close = False

        pcount = self.header["num_halos"]
        pos = np.empty((pcount, 3), dtype="float64")
        for i, ax in enumerate("xyz"):
            pos[:, i] = f[f"particle_position_{ax}"][()]

        if close:
            f.close()

        return pos


class YTHaloCatalogDataset(SavedDataset):
    """
    Dataset class for halo catalogs made with yt.

    This covers yt FoF/HoP halo finders and the halo analysis
    in yt_astro_analysis.
    """

    _load_requirements = ["h5py"]
    _index_class = ParticleIndex
    _file_class = YTHaloCatalogFile
    _field_info_class = YTHaloCatalogFieldInfo
    _suffix = ".h5"
    _con_attrs = (
        "cosmological_simulation",
        "current_time",
        "current_redshift",
        "hubble_constant",
        "omega_matter",
        "omega_lambda",
        "domain_left_edge",
        "domain_right_edge",
    )

    def __init__(
        self,
        filename,
        dataset_type="ythalocatalog",
        index_order=None,
        units_override=None,
        unit_system="cgs",
    ):
        self.index_order = index_order
        super().__init__(
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )

    def add_field(self, *args, **kwargs):
        super().add_field(*args, **kwargs)
        self._halos_ds.add_field(*args, **kwargs)

    @property
    def halos_field_list(self):
        return self._halos_ds.field_list

    @property
    def halos_derived_field_list(self):
        return self._halos_ds.derived_field_list

    @cached_property
    def _halos_ds(self):
        return YTHaloDataset(self)

    def _setup_classes(self):
        super()._setup_classes()
        self.halo = partial(YTHaloCatalogHaloContainer, ds=self._halos_ds)
        self.halo.__doc__ = YTHaloCatalogHaloContainer.__doc__

    def _parse_parameter_file(self):
        self.refine_by = 2
        self.dimensionality = 3
        self.domain_dimensions = np.ones(self.dimensionality, "int32")
        self._periodicity = (True, True, True)
        prefix = ".".join(self.parameter_filename.rsplit(".", 2)[:-2])
        self.filename_template = f"{prefix}.%(num)s{self._suffix}"
        self.file_count = len(glob.glob(prefix + "*" + self._suffix))
        self.particle_types = ("halos",)
        self.particle_types_raw = ("halos",)
        super()._parse_parameter_file()

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if not filename.endswith(".h5"):
            return False

        if cls._missing_load_requirements():
            return False

        with h5py.File(filename, mode="r") as f:
            if (
                "data_type" in f.attrs
                and parse_h5_attr(f, "data_type") == "halo_catalog"
            ):
                return True
        return False


class YTHaloParticleIndex(ParticleIndex):
    """
    Particle index for getting halo particles from YTHaloCatalogDatasets.
    """

    def __init__(self, ds, dataset_type):
        self.real_ds = weakref.proxy(ds.real_ds)
        super().__init__(ds, dataset_type)

    def _calculate_particle_index_starts(self):
        """
        Create a dict of halo id offsets for each file.
        """
        particle_count = defaultdict(int)
        offset_count = 0
        for data_file in self.data_files:
            data_file.index_start = {
                ptype: particle_count[ptype] for ptype in data_file.total_particles
            }
            data_file.offset_start = offset_count
            for ptype in data_file.total_particles:
                particle_count[ptype] += data_file.total_particles[ptype]
            offset_count += getattr(data_file, "total_offset", 0)

        self._halo_index_start = {}
        for ptype in self.ds.particle_types_raw:
            d = [data_file.index_start[ptype] for data_file in self.data_files]
            self._halo_index_start.update({ptype: np.array(d)})

    def _detect_output_fields(self):
        field_list = []
        scalar_field_list = []
        units = {}

        pc = {}
        for ptype in self.ds.particle_types_raw:
            d = [df.total_particles[ptype] for df in self.data_files]
            pc.update({ptype: sum(d)})
        found_fields = {ptype: False for ptype, pnum in pc.items() if pnum > 0}
        has_ids = False

        for data_file in self.data_files:
            fl, sl, idl, _units = self.io._identify_fields(data_file)
            units.update(_units)
            field_list.extend([f for f in fl if f not in field_list])
            scalar_field_list.extend([f for f in sl if f not in scalar_field_list])
            for ptype in found_fields:
                found_fields[ptype] |= data_file.total_particles[ptype]
            has_ids |= len(idl) > 0
            if all(found_fields.values()) and has_ids:
                break

        self.field_list = field_list
        self.scalar_field_list = scalar_field_list
        ds = self.dataset
        ds.scalar_field_list = scalar_field_list
        ds.particle_types = tuple({pt for pt, ds in field_list})
        ds.field_units.update(units)
        ds.particle_types_raw = ds.particle_types

    def _get_halo_file_indices(self, ptype, identifiers):
        """
        Get the index of the data file list where this halo lives.

        Digitize returns i such that bins[i-1] <= x < bins[i], so we subtract
        one because we will open data file i.
        """
        return np.digitize(identifiers, self._halo_index_start[ptype], right=False) - 1

    def _get_halo_scalar_index(self, ptype, identifier):
        i_scalar = self._get_halo_file_indices(ptype, [identifier])[0]
        scalar_index = identifier - self._halo_index_start[ptype][i_scalar]
        return scalar_index

    def _get_halo_values(self, ptype, identifiers, fields, f=None):
        """
        Get field values for halo data containers.
        """

        # if a file is already open, don't open it again
        filename = None if f is None else f.filename

        data = defaultdict(lambda: np.empty(identifiers.size))
        i_scalars = self._get_halo_file_indices(ptype, identifiers)
        for i_scalar in np.unique(i_scalars):
            # mask array to get field data for this halo
            target = i_scalars == i_scalar
            scalar_indices = identifiers - self._halo_index_start[ptype][i_scalar]

            # only open file if it's not already open
            my_f = (
                f
                if self.data_files[i_scalar].filename == filename
                else h5py.File(self.data_files[i_scalar].filename, mode="r")
            )

            for field in fields:
                data[field][target] = self._read_halo_particle_field(
                    my_f, ptype, field, scalar_indices[target]
                )

            if self.data_files[i_scalar].filename != filename:
                my_f.close()

        return data

    def _identify_base_chunk(self, dobj):
        pass

    def _read_halo_particle_field(self, fh, ptype, field, indices):
        return fh[field][indices]

    def _read_particle_fields(self, fields, dobj, chunk=None):
        if not fields:
            return {}, []
        fields_to_read, fields_to_generate = self._split_fields(fields)
        if not fields_to_read:
            return {}, fields_to_generate
        fields_to_return = self.io._read_particle_selection(dobj, fields_to_read)
        return fields_to_return, fields_to_generate

    def _setup_data_io(self):
        super()._setup_data_io()
        if self.real_ds._instantiated_index is None:
            self.real_ds.index
        self.real_ds.index

        # inherit some things from parent index
        self._data_files = self.real_ds.index.data_files
        self._total_particles = self.real_ds.index.total_particles

        self._calculate_particle_index_starts()


class HaloDataset(ParticleDataset):
    """
    Base class for dataset accessing particles from halo catalogs.
    """

    def __init__(self, ds, dataset_type):
        self.real_ds = ds
        for attr in [
            "filename_template",
            "file_count",
            "particle_types_raw",
            "particle_types",
            "_periodicity",
        ]:
            setattr(self, attr, getattr(self.real_ds, attr))

        super().__init__(self.real_ds.parameter_filename, dataset_type)

    def print_key_parameters(self):
        pass

    def _set_derived_attrs(self):
        pass

    def _parse_parameter_file(self):
        for attr in [
            "cosmological_simulation",
            "cosmology",
            "current_redshift",
            "current_time",
            "dimensionality",
            "domain_dimensions",
            "domain_left_edge",
            "domain_right_edge",
            "domain_width",
            "hubble_constant",
            "omega_lambda",
            "omega_matter",
            "unique_identifier",
        ]:
            setattr(self, attr, getattr(self.real_ds, attr, None))

    def set_code_units(self):
        self._set_code_unit_attributes()
        self.unit_registry = self.real_ds.unit_registry

    def _set_code_unit_attributes(self):
        for unit in ["length", "time", "mass", "velocity", "magnetic", "temperature"]:
            my_unit = f"{unit}_unit"
            setattr(self, my_unit, getattr(self.real_ds, my_unit, None))

    def __str__(self):
        return f"{self.real_ds}"

    def _setup_classes(self):
        self.objects = []


class YTHaloDataset(HaloDataset):
    """
    Dataset used for accessing member particles from YTHaloCatalogDatasets.
    """

    _index_class = YTHaloParticleIndex
    _file_class = YTHaloCatalogFile
    _field_info_class = YTHaloCatalogHaloFieldInfo

    def __init__(self, ds, dataset_type="ythalo"):
        super().__init__(ds, dataset_type)

    def _set_code_unit_attributes(self):
        pass

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        # We don't ever want this to be loaded by yt.load.
        return False


class HaloContainer(YTSelectionContainer):
    """
    Base class for data containers providing halo particles.
    """

    _type_name = "halo"
    _con_args = ("ptype", "particle_identifier")
    _skip_add = True
    _spatial = False

    def __init__(self, ptype, particle_identifier, ds=None):
        if ptype not in ds.particle_types_raw:
            raise RuntimeError(
                f'Possible halo types are {ds.particle_types_raw}, supplied "{ptype}".'
            )

        self.ptype = ptype
        self._current_particle_type = ptype
        super().__init__(ds, {})

        self._set_identifiers(particle_identifier)

        # Find the file that has the scalar values for this halo.
        i_scalar = self.index._get_halo_file_indices(ptype, [self.particle_identifier])[
            0
        ]
        self.i_scalar = i_scalar
        self.scalar_data_file = self.index.data_files[i_scalar]

        # Data files containing particles belonging to this halo.
        self.field_data_files = [self.index.data_files[i_scalar]]

        # index within halo arrays that corresponds to this halo
        self.scalar_index = self.index._get_halo_scalar_index(
            ptype, self.particle_identifier
        )

        self._set_io_data()
        self.particle_number = self._get_particle_number()

        # starting and ending indices for each file containing particles
        self._set_field_indices()

    @cached_property
    def mass(self):
        return self[self.ptype, "particle_mass"][0]

    @cached_property
    def radius(self):
        return self[self.ptype, "virial_radius"][0]

    @cached_property
    def position(self):
        return self[self.ptype, "particle_position"][0]

    @cached_property
    def velocity(self):
        return self[self.ptype, "particle_velocity"][0]

    def _set_io_data(self):
        halo_fields = self._get_member_fieldnames()
        my_data = self.index._get_halo_values(
            self.ptype, np.array([self.particle_identifier]), halo_fields
        )
        self._io_data = {field: np.int64(val[0]) for field, val in my_data.items()}

    def __repr__(self):
        return f"{self.ds}_{self.ptype}_{self.particle_identifier:09d}"


class YTHaloCatalogHaloContainer(HaloContainer):
    """
    Data container for accessing particles from a halo.

    Create a data container to get member particles and individual
    values from halos and subhalos. Halo mass, radius, position, and
    velocity are set as attributes. Halo IDs are accessible
    through the field, "member_ids".  Other fields that are one
    value per halo are accessible as normal.  The field list for
    halo objects can be seen in `ds.halos_field_list`.

    Parameters
    ----------
    ptype : string
        The type of halo. Possible options can be found by
        inspecting the value of ds.particle_types_raw.
    particle_identifier : int
        The halo id.

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("tiny_fof_halos/DD0046/DD0046.0.h5")

    >>> halo = ds.halo("halos", 0)
    >>> print(halo.particle_identifier)
    0
    >>> print(halo.mass)
    8724990744704.453 Msun
    >>> print(halo.radius)
    658.8140635766607 kpc
    >>> print(halo.position)
    [0.05496909 0.19451951 0.04056824] code_length
    >>> print(halo.velocity)
    [7034181.07118151 5323471.09102874 3234522.50495914] cm/s
    >>> # particle ids for this halo
    >>> print(halo["member_ids"])
    [ 1248.   129.   128. 31999. 31969. 31933. 31934.   159. 31903. 31841. ...
      2241.  2240.  2239.  2177.  2209.  2207.  2208.] dimensionless

    """

    def _get_member_fieldnames(self):
        return ["particle_number", "particle_index_start"]

    def _get_particle_number(self):
        return self._io_data["particle_number"]

    def _set_field_indices(self):
        self.field_data_start = [self._io_data["particle_index_start"]]
        self.field_data_end = [self.field_data_start[0] + self.particle_number]

    def _set_identifiers(self, particle_identifier):
        self.particle_identifier = particle_identifier
        self.group_identifier = self.particle_identifier
